"""Machinery to read/write Tables in an Excel workbook using openpyxl as engine."""
from typing import Union, Iterable, Sequence, Any
from os import PathLike

import openpyxl

try:
    from openpyxl.worksheet.worksheet import Worksheet as OpenpyxlWorksheet
except ImportError:
    # openpyxl < 2.6
    from openpyxl.worksheet import Worksheet as OpenpyxlWorksheet


from pdtable import Table
from pdtable.io._represent import _represent_row_elements


def read_cell_rows_openpyxl(path: Union[str, PathLike]) -> Iterable[Sequence[Any]]:
    """Reads from an Excel workbook, yielding one row of cells at a time."""
    import openpyxl

    wb = openpyxl.load_workbook(path)
    for ws in wb.worksheets:
        yield from ws.iter_rows(values_only=True)


def write_excel_openpyxl(tables, path, na_rep):
    """Writes tables to an Excel workbook at the specified path."""
    if isinstance(tables, Table):
        # For convenience, pack single table in an iterable
        tables = [tables]
    wb = openpyxl.Workbook()
    ws = wb.active
    for t in tables:
        _append_table_to_openpyxl_worksheet(t, ws, na_rep)
    wb.save(path)


def _append_table_to_openpyxl_worksheet(
    table: Table, ws: OpenpyxlWorksheet, na_rep: str = "-"
) -> None:
    units = table.units
    ws.append([f"**{table.name}"])
    ws.append([" ".join(str(x) for x in table.metadata.destinations)])
    ws.append(table.column_names)
    ws.append(units)
    for row in table.df.itertuples(index=False, name=None):
        # TODO: apply format string specified in ColumnMetadata
        ws.append(_represent_row_elements(row, units, na_rep))
    ws.append([])  # blank row marking table end



"""
The `frame` module allows working with StarTable tables as pandas dataframes.

This is implemented by providing both `Table` and `TableDataFrame` interfaces to the same object.

## Idea

The central idea is that as much as possible of the table information is stored as a pandas
dataframe, and that the remaining information is stored as a `ComplementaryTableInfo` object
attached to the dataframe as registered metadata. Further, access to the full table data
structure is provided through a facade object (of class `Table`). `Table` objects have no state
(except the underlying decorated dataframe) and are intended to be created when needed
and discarded afterwards:

```
dft = make_table_dataframe(...)
unit_height = Table(dft).height.unit
```

Advantages of this approach are that:

1. Code can be written for (and tested with) pandas dataframes and still operate on `TableDataFrame`
   objects. This avoids unnecessary coupling to the StarTable project.
2. The table access methods of pandas are available for use by consumer code. This both saves
   the work of making startable-specific access methods, and likely allows better performance
   and documentation.

## Implementation details

The decorated dataframe objects are represented by the `TableDataFrame` class.

### Dataframe operations

Pandas allows us to hook into operations on dataframes via the `__finalize__` method.
This makes it possible to propagate table metadata over select dataframe operations.
See `TableDataFrame` documentation for details.

### Column metadata

Propagating metadata would be greatly simplified if column-specific metadata was stored with the
column. However, metadata attached to `Series` object is not retained within the dataframe, see
https://github.com/pandas-dev/pandas/issues/6923#issuecomment-41043225.

## Alternative approaches

It should be possible to maintain column metadata together with the column data through the use of
`ExtensionArray`. This option was discarded early due to performance concerns, but might be viable
and would the be preferable to the chosen approach.
"""

import pandas as pd
import warnings
from typing import Set, Dict, Optional, Iterable

from .table_metadata import TableMetadata, ColumnMetadata, ComplementaryTableInfo

_TABLE_INFO_FIELD_NAME = "_table_data"


class UnknownOperationError(Exception):
    pass


class InvalidTableCombineError(Exception):
    pass


def _combine_tables(obj: "TableDataFrame", other, method, **kwargs) -> ComplementaryTableInfo:
    """
    Called from __finalize__ when operations have been performed via the pandas.DataFrame API.

    Implementation policy is that this will fail except for situations where
    the metadata combination is safe.
    For other cases, the operations should be implemented via the Table facade
    """

    if method is None:
        # slicing
        src = [other]
    elif method == "merge":
        src = [other.left, other.right]
    elif method == "concat":
        src = other.objs
    elif method == "copy":
        src = [other]  # TODO should it be obj?  Test it
    else:
        raise UnknownOperationError(
            f"Unknown method while combining metadata: {method}. Keyword args: {kwargs}"
        )

    if len(src) == 0:
        raise UnknownOperationError(f"No operands for operation {method}")

    data = [get_table_info(s) for s in src if is_table_dataframe(s)]

    # 1: Create table metadata as combination of all
    meta = TableMetadata(
        name=data[0].metadata.name, operation=f"Pandas {method}", parents=[d.metadata for d in data]
    )

    # 2: Check that units match for columns that appear in more than one table
    out_cols: Set[str] = set(obj.columns)
    columns: Dict[str, ColumnMetadata] = {}
    for d in data:
        for name, c in d.columns.items():
            if name not in out_cols:
                continue
            col = columns.get(name, None)
            if not col:
                # not seen before in input
                col = c.copy()
                columns[name] = col
            else:
                if not col.unit == c.unit:
                    raise InvalidTableCombineError(
                        "Column {name} appears with incompatible units.", col.unit, c.unit
                    )
                col.update_from(c)

    return ComplementaryTableInfo(table_metadata=meta, columns=columns)


class TableDataFrame(pd.DataFrame):
    """
    A pandas.DataFrame subclass with associated table metadata.

    Behaves exactly as a pandas.DataFrame, and will try to retain metadata
    through pandas operations. If this is not possible, the manipulations
    will return a plain pandas.DataFrame.

    No StarTable-specific methods are available directly on this class, and TableDataFrame
    objects should not be created directly.
    Instead, use either the methods in the this module, or the Table proxy
    object, which can be constructed for a TableDataFrame object 'tdf' via Table(tdf).
    """

    _metadata = [_TABLE_INFO_FIELD_NAME]  # Register metadata fieldnames here

    # If implemented, must handle metadata copying etc
    # def __init__(self, *args, **kwargs):
    #    super().__init__(*args, **kwargs)

    # These are implicit by inheritance from dataframe
    # We could override _constructor_sliced to add metadata to exported columns
    # but not clear if this would add value
    #     @property
    #     def _constructor_expanddim(self):
    #         return pd.DataFrame
    #     @property
    #     def _constructor_sliced(self):
    #         return pd.Series

    @property
    def _constructor(self):
        return TableDataFrame

    def __finalize__(self, other, method=None, **kwargs):
        """
        Overrides pandas.core.generic.NDFrame.__finalize__()

        This method is responsible for populating TableDataFrame metadata
        when creating a new Table object.

        If left out, no metadata would be retained across table
        operations. This might be a viable solution?
        Alternatively, we could return a raw frame object on some operations
        """

        try:
            data = _combine_tables(self, other, method, **kwargs)
            object.__setattr__(self, _TABLE_INFO_FIELD_NAME, data)
        except UnknownOperationError as e:
            warnings.warn(f"Falling back to dataframe: {e}")
            return pd.DataFrame(self)
        return self

    @staticmethod
    def from_table_info(df: pd.DataFrame, table_info: ComplementaryTableInfo) -> "TableDataFrame":
        df = TableDataFrame(df)
        object.__setattr__(df, _TABLE_INFO_FIELD_NAME, table_info)
        table_info._check_dataframe(df)
        return df


def is_table_dataframe(df: pd.DataFrame) -> bool:
    return _TABLE_INFO_FIELD_NAME in df._metadata


def make_table_dataframe(
    df: pd.DataFrame,
    units: Optional[Iterable[str]] = None,
    unit_map: Optional[Dict[str, str]] = None,
    table_metadata: Optional[TableMetadata] = None,
    **kwargs,
) -> TableDataFrame:
    """
    Create TableDataFrame object from a pandas.DataFream and table metadata elements.

    Unknown keyword arguments (e.g. `name = ...`) are used to create a `TableMetadata` object.
    Alternatively, a `TableMetadata` object can be provided directly.

    Either a list of units for all columns, or a unit_map can be provided. Otherwise, default units
    are assigned.

    Example:
    tdf = make_table_dataframe(df, name='MyTable')
    """

    # build metadata
    if (table_metadata is not None) == bool(kwargs):
        raise Exception("Supply either metadata or keyword-arguments for TableMetadata constructor")
    if kwargs:
        # This is intended to fail if args are insufficient
        table_metadata = TableMetadata(**kwargs)

    df = TableDataFrame.from_table_info(
        df, table_info=ComplementaryTableInfo(table_metadata=table_metadata)
    )

    # set units
    if units and unit_map:
        raise Exception("Supply at most one of unit and unit_map")
    if units is not None:
        set_all_units(df, units)
    elif unit_map is not None:
        set_units(df, unit_map)

    return df


def get_table_info(
    df: TableDataFrame, fail_if_missing=True, check_dataframe=True
) -> Optional[ComplementaryTableInfo]:
    """
    Get ComplementaryTableInfo from existing TableDataFrame object.

    When called with default options, get_table_info will either raise an exception
    or return a ComplementaryTableInfo object with a valid ColumnMetadata defined for each column.

    check_dataframe: Check that the table data is valid with respect to dataframe.
                     If the dataframe has been manipulated directly, table will be updated to match.
    fail_if_missing: Whether to raise an exception if ComplementaryTableInfo object is missing
    """
    name: str = _TABLE_INFO_FIELD_NAME
    if name not in df._metadata:
        raise Exception(
            "Attempt to extract table data from normal pd.DataFrame object."
            "ComplementaryTableInfo can only be associated with TableDataFrame objects"
        )
    table_data = getattr(df, _TABLE_INFO_FIELD_NAME, None)
    if not table_data:
        if fail_if_missing:
            raise Exception(
                "Missing ComplementaryTableInfo object on TableDataFrame. TableDataFrame objects "
                "should be created via make_table_dataframe or a Table proxy."
            )
    elif check_dataframe:
        table_data._check_dataframe(df)
    return table_data


# example of manipulator function that directly manipulates TableDataFrame objects without constructing Table facade  # noqa
def add_column(df: TableDataFrame, name: str, values, unit: Optional[str] = None, **kwargs):
    """
    Add or update column in table. If omitted, unit will be partially inferred from value dtype.

    keyword arguments will be forwarded to ColumnMetadata constructor together with unit
    """
    df[name] = values
    columns = get_table_info(df, check_dataframe=False).columns

    new_col = (
        ColumnMetadata.from_dtype(df[name].dtype, **kwargs)
        if unit is None
        else ColumnMetadata(unit=unit, **kwargs)
    )

    col = columns.get(name, None)
    if col is None:
        columns[name] = new_col
    else:
        # Existing column
        col.update_from(new_col)


def set_units(df: TableDataFrame, unit_map: Dict[str, str]):
    columns = get_table_info(df).columns
    for col, unit in unit_map.items():
        columns[col] = unit


def set_all_units(df: TableDataFrame, units: Iterable[Optional[str]]):
    """
    Set units for all columns in table.
    """
    columns = get_table_info(df).columns
    for col, unit in zip(df.columns, units):
        columns[col].unit = unit



import io
from textwrap import dedent
from typing import List
from pathlib import Path

from pytest import fixture, raises

from pdtable import read_csv, BlockType, Table


@fixture
def csv_data() -> str:
    return dedent(
        """\
        author: ;XYODA     ;
        purpose:;Save the galaxy;

        ***gunk
        grok
        jiggyjag

        **places;
        all
        place;distance;ETA;is_hot
        text;km;datetime;onoff
        home;0.0;2020-08-04 08:00:00;1
        work;1.0;2020-08-04 09:00:00;0
        beach;2.0;2020-08-04 17:00:00;1

        ::;details about various places;;

        **farm_animals;;;
        your_farm my_farm other_farm;;;
        species;n_legs;avg_weight;
        text;-;kg;
        chicken;2;2;
        pig;4;89;
        cow;4;200;
        unicorn;4;NaN;
        """
    )


def test_read_csv(csv_data):
    bl = list(read_csv(io.StringIO(csv_data)))
    tables: List[Table] = [b for t, b in bl if t == BlockType.TABLE]
    template_rows = [b for t, b in bl if t == BlockType.TEMPLATE_ROW]
    met = [b for t, b in bl if t == BlockType.METADATA]

    assert len(met) == 1
    assert tables[0].df["place"][1] == "work"
    assert len(template_rows) == 1


def test_read_csv__sep_is_comma(csv_data):
    bl = list(read_csv(io.StringIO(csv_data.replace(";", ",")), sep=","))
    tables: List[Table] = [b for t, b in bl if t == BlockType.TABLE]
    template_rows = [b for t, b in bl if t == BlockType.TEMPLATE_ROW]
    met = [b for t, b in bl if t == BlockType.METADATA]

    assert len(met) == 1
    assert tables[0].df["place"][1] == "work"
    assert len(template_rows) == 1


def test_read_csv__from_stream():
    with open(Path(__file__).parent / "input" / "bundle.csv", "r") as fh:
        bls = list(read_csv(fh))
        tables = [bl for ty, bl in bls if ty == BlockType.TABLE]
        assert tables[1].name == "spelling_numbers"

    # raises exception on common error if not text stream
    with raises(Exception):
        with open(Path(__file__).parent / "input" / "bundle.csv", "rb") as fh:  # binary stream!
            bls = list(read_csv(fh))
            tables = [bl for ty, bl in bls if ty == BlockType.TABLE]



"""Interface to read/write StarTable data from/to CSV"""
import os
import io
from contextlib import nullcontext
from os import PathLike

from typing import TextIO, Union, Callable, Iterable

import pdtable  # Required to read dynamically-set pdtable.CSV_SEP
from ._represent import _represent_row_elements
from .. import BlockType, Table, TableBundle
from ..store import BlockIterator
from .parsers.fixer import ParseFixer
from .parsers.blocks import parse_blocks


def read_csv(
    source: Union[str, PathLike, TextIO],
    sep: str = None,
    origin: str = None,
    fixer: ParseFixer = None,
    to: str = "pdtable",
    filter: Callable[[BlockType, str], bool] = None,
) -> BlockIterator:
    """Reads StarTable data from a CSV file or text stream, yielding one block at a time.

    Reads StarTable data from a CSV-format file or text stream. StarTable blocks are parsed from
    this data and yielded one at a time as a (block_type, block_content) tuple, where
    - 'block_type' is a BlockType enum indicating which of the StarTable block types this is (table,
      metadata, directive, template); and
    - 'block_content' is the block content.

    'block_content' is presented as one of the following data types, depending on the 'to' argument
    passed:
    - A pdtable.Table object,
    - A JSON serializable object (structure of nested dicts and lists of JSON-mappable values); or
    - A list of list of values, representing the raw cell grid (row and columns from the CSV data).

    Blocks can be filtered prior to parsing, by passing a callable as 'filter' argument. This can
    reduce reading time substantially when reading a subset of tables from an otherwise large file
    or stream. Only those blocks for which 'filter' returns True are fully parsed. Other blocks
    are parsed only superficially i.e. only the block's top-left cell, which is just
    enough to recognize block type and name to pass to 'filter', thus avoiding the much more
    expensive task of parsing the entire block, e.g. the values in all columns and rows of a large
    table.

    This is a thin wrapper around parse_blocks(). The only thing it does is present the contents of
    a CSV file or stream as a Iterable of cell rows, where each row is a sequence of values.

    Args:
        source:
            File path or text stream from which to read.
            If a file path, then this file gets opened, and then closed after reading.
            If a stream, then it is left open after reading; the caller is responsible for managing
            the stream.

        sep:
            Optional; CSV field delimiter. Default is ';'.

        origin:
            Optional; Table location

        fixer:
            Customized ParseFixer instance to be used instead of default fixer.
            fixer corrects simple errors in source stream.

        to:
            Determines the data type of the yielded blocks. Can be either of:
            - 'pdtable' (Default): pdtable.Table and other pdtable-style block objects
            - 'jsondata': JSON serializable objects (nested structure of dicts, lists, and
              JSON-mappable values) that can be passed directly to json.dump()
            - 'cellgrid': A grid of raw input cells i.e. a List[List[values]]

        filter:
            A callable that takes a (BlockType, block_name) tuple, and returns true if a block
            meeting this description is to be parsed, or false if it is to be ignored and discarded.

    Yields:
        Tuples of (BlockType, block) where 'block' is one of {Table, MetadataBlock, Directive,
        TemplateBlock}

    """
    if sep is None:
        sep = pdtable.CSV_SEP

    if origin is None:
        if hasattr(source, "name"):
            origin = source.name
        else:
            origin = str(source)

    kwargs = {"sep": sep, "origin": origin, "fixer": fixer, "to": to, "filter": filter}

    if not isinstance(source, (str, PathLike)):
        assert isinstance(source, io.TextIOBase)

    with open(source) if isinstance(source, (str, PathLike)) else nullcontext(source) as f:
        cell_rows = (line.rstrip("\n").split(sep) for line in f)
        yield from parse_blocks(cell_rows, **kwargs)


def write_csv(
    tables: Union[Table, Iterable[Table], TableBundle],
    to: Union[str, os.PathLike, TextIO],
    sep: str = None,
    na_rep: str = "-",
):
    """Writes one or more tables to a CSV file or text stream.

    Writes table blocks in CSV format to a file or text stream. Values are formatted to comply with
    the StarTable standard where necessary and possible; otherwise they are simply str()'ed.

    Args:
        tables:
            Table(s) to write. Can be a single Table or an iterable of Tables.
        to:
            File path or text stream to which to write.
            If a file path, then this file gets created/overwritten and then closed after writing.
            If a stream, then it is left open after writing; the caller is responsible for managing
            the stream.
        sep:
            Optional; CSV field delimiter. Default is ';'.
        na_rep:
            Optional; String representation of missing values (NaN, None, NaT). Default is '-'.
            If overriding this default, use another value compliant with the StarTable standard.
    """
    if sep is None:
        sep = pdtable.CSV_SEP

    if isinstance(tables, Table):
        # For convenience, pack single table in an iterable
        tables = [tables]

    # If it looks like a path, open a file and close when done.
    # Else we assume it's a stream that the caller is responsible for managing; leave it open.
    with open(to, "w") if isinstance(to, (str, os.PathLike)) else nullcontext(to) as stream:
        for table in tables:
            _table_to_csv(table, stream, sep, na_rep)


def _table_to_csv(table: Table, stream: TextIO, sep: str, na_rep: str) -> None:
    """Writes a single Table to stream as CSV.
    """

    units = table.units
    display_formats = [table.column_metadata[c].display_format for c in table.column_metadata]
    format_strings = [f"{{:{f.specifier}}}" if f else None for f in display_formats]

    # Build entire string at once
    the_whole_thing = (
        f"**{table.name}{sep}\n"
        + " ".join(str(x) for x in table.metadata.destinations)
        + "\n"
        + sep.join(str(x) for x in table.column_names)
        + "\n"
        + sep.join(str(x) for x in units)
        + "\n"
        + "\n".join(
            sep.join(
                fs.format(x) if fs else str(x)
                for x, fs in zip(_represent_row_elements(row, units, na_rep), format_strings)
            )
            for row in table.df.itertuples(index=False, name=None)
        )
        + "\n\n"
    )

    stream.write(the_whole_thing)



import json
import os
from pathlib import Path

import pandas as pd
import pytest

from pdtable import ParseFixer, BlockType
from pdtable import read_csv
from pdtable.io import table_to_json_data
from pdtable.io.parsers import parse_blocks
from pdtable.io.parsers.blocks import make_table


class custom_test_fixer(ParseFixer):
    def __init__(self):
        ParseFixer.__init__(self)
        self.stop_on_errors = False
        self._called_from_test = True


def input_dir() -> Path:
    return Path(__file__).parent / "input/with_errors"


def test_columns_duplicate():
    """
       Verify that default ParseFixer corrects duplicate column names

    """
    tab = None
    with open(input_dir() / "cols1.csv", "r") as fh:
        g = read_csv(fh, fixer=custom_test_fixer)
        for tp, tt in g:
            if True:
                if tp == BlockType.TABLE:
                    tab = tt
                    break
    assert tab is not None
    assert tab.df["flt_fixed_001"] is not None
    assert tab.df["flt_fixed_001"][6] == 7.6
    assert tab.df["flt"][0] == 3.0


def test_columns_missing():
    """
       Verify that default ParseFixer corrects missing column name

    """
    tab = None
    with open(input_dir() / "cols2.csv", "r") as fh:
        g = read_csv(fh, fixer=custom_test_fixer)
        for tp, tt in g:
            if True:
                if tp == BlockType.TABLE:
                    tab = tt
                    break
    assert tab is not None
    assert tab.df["missing_fixed_000"] is not None
    assert tab.df["flt"][6] == 7.11


def test_custom_fixer():
    """ Test custom ParseFixer
        Verify that read_csv uses custom ParseFixer
    """

    class fix_pi(ParseFixer):
        def __init__(self):
            super().__init__()

        # augment existing method, simple fix float
        def fix_illegal_cell_value(self, vtype, value):
            if vtype == "float":
                return 22.0 / 7.0
            else:
                fix_value = ParseFixer.fix_illegal_cell_value(self, vtype, value)
                return fix_value

    fix = fix_pi()
    fix.stop_on_errors = False
    fix._called_from_test = True

    with open(input_dir() / "types3.csv", "r") as fh:
        g = read_csv(fh, to="jsondata", fixer=fix)
        for tp, tt in g:
            if tp == BlockType.TABLE:
                assert tt["columns"]["num"]["values"][2] == 22.0 / 7.0
                assert tt["columns"]["flt"]["values"][0] == 22.0 / 7.0
                assert tt["columns"]["flt"]["values"][0] == 22.0 / 7.0
                assert tt["columns"]["flt2"]["values"][2] == 22.0 / 7.0

    with pytest.raises(ValueError):
        # test read_csv w. class (not instance) of fixer
        # class has default stop_on_errors = True
        with open(input_dir() / "types3.csv", "r") as fh:
            g = read_csv(fh, to="jsondata", fixer=fix_pi)
            for tp, tt in g:
                pass


def test_FAT():
    """ Factory Acceptance Test

        Verify that we are able to read all files in ./input
        Using default ParseFixer
    """
    all_files = 0
    ignore_files = ["auto_fixed.py", "__init__.py", "all.json"]
    for fn in os.listdir(input_dir()):
        path = input_dir() / fn
        if not os.path.isfile(path):
            continue
        if fn in ignore_files:
            continue
        all_files += 1

    # load targets
    with open(input_dir() / "all.json") as f:
        all_json = json.load(f)

    for fn in os.listdir(input_dir()):
        path = input_dir() / fn
        if not os.path.isfile(path):
            continue
        if fn in ignore_files:
            continue

        with open(input_dir() / fn, "r") as fh:
            g = read_csv(fh, origin=f'"{fn}"', to="jsondata", fixer=custom_test_fixer)
            count = 0
            for tp, tt in g:
                if tp == BlockType.TABLE:
                    count += 1
                    if fn != "all.csv":
                        assert tt == all_json[fn]

            if fn == "all.csv":
                assert count == all_files - 1
            else:
                assert count == 1


def test_stop_on_errors():
    """ Unit test cusrom ParseFixer.stop_on_errors
    """
    # fmt: off
    table_lines = [
        ["**tab_ok"],
        ["dst1"],
        [ "a1"  , "a2"  , "a3"  , "a4"  ],
        [ "-"   , "-"   , "-"   , "-"   ],
        [ 1     , 2     , 3     , 3.14  ],
        [],
        ["**tab_errors"],
        ["dst1"],
        [ "a1"  , "a2"  , "a3"  , "a4"  ],
        [ "-"   , "-"   , "-"   , "-"   ],
        [ "NaN" , "nan" , "Nine", "Ten" ],
        [ 1     , 2     , 3     , 4     ],
    ]
    # fmt: on

    fix = ParseFixer()
    fix.stop_on_errors = True
    fix._dbg = False  # ignore during test
    pi = 0
    with pytest.raises(ValueError):
        for typ, tab in parse_blocks(table_lines, fixer=fix, to="pdtable"):
            if typ != BlockType.TABLE:
                continue
            assert tab.df["a4"][0] == 3.14
            pi += 1

    with pytest.raises(ValueError):
        for typ, tab in parse_blocks(table_lines, fixer=fix, to="jsondata"):
            if typ != BlockType.TABLE:
                continue
            assert tab["columns"]["a4"]["values"][0] == 3.14
            pi += 1

    # cellgrid does not parse values, i.e. no ValueError
    for typ, tab in parse_blocks(table_lines, fixer=fix, to="cellgrid"):
        if typ != BlockType.TABLE:
            continue
        if tab[0][0] == "**tab_ok":
            assert tab[4][3] == 3.14
            pi += 1
        if tab[0][0] == "**tab_errors":
            assert tab[4][3] == "Ten"

    assert pi == 3  # 😉


def test_stop_on_errors_default_fixer():
    """ Unit test ParseFixer: raise ValueError on empty float and empty onoff
    """
    # fmt: off
    table_lines_flt = [
        ["**tab_ok"],
        ["dst1"],
        [ "a1"  , "a2"  , "a3"  , "a4"  ],
        [ "-"   , "-"   , "-"   , "-"   ],
        [ 1     , 2     , 3     , 3.14  ],
        [],
        ["**tab_errors"],
        ["dst1"],
        [ "a1"  , "a2"  , "a3"  , "a4"  ],
        [ "-"   , "-"   , "-"   , "-"   ],
        [ 1     , 2     , 3     , ""    ],
    ]
    # fmt: on

    g = parse_blocks(table_lines_flt, filter=lambda ty, tn: ty == BlockType.TABLE)
    typ, tab = next(g)
    assert tab.name == "tab_ok"
    assert tab.df["a4"][0] == 3.14

    with pytest.raises(ValueError):
        typ, tab = next(g)
        print(f"-oOo-: {typ} {tab}")

    # fmt: off
    table_lines_onoff = [
        ["**tab_ok"],
        ["dst1"],
        [ "a1"  , "a2"],
        [ "-"   , "onoff"],
        [ 3.14  , 1      ],
        [],
        ["**tab_errors"],
        ["dst1"],
        [ "a1"  , "a2" ],
        [ "-"   , "onoff" ],
        [ 1     ,  None   ],
    ]
    # fmt: on
    g = parse_blocks(table_lines_onoff, filter=lambda ty, tn: ty == BlockType.TABLE)
    typ, tab = next(g)
    assert tab.name == "tab_ok"
    assert tab.df["a1"][0] == 3.14

    with pytest.raises(ValueError):
        typ, tab = next(g)
        print(f"-oOo-: {typ} {tab}")

    # fmt: off
    table_lines_datetime = [
        ["**tab_ok"],
        ["dst1"],
        [ "a1"  , "a2"],
        [ "-"   , "datetime"   ],
        [ 14    , pd.to_datetime("2020-08-11")],
        [ 3.14  , "2020-08-12" ],
        [],
        ["**tab_errors"],
        ["dst1"],
        [ "a1"  , "a2" ],
        [ "-"   , "datetime" ],
        [ 14    , pd.to_datetime("2020-08-13")],
        [ 1     ,  None   ],
    ]
    # fmt: on
    g = parse_blocks(table_lines_datetime, filter=lambda ty, tn: ty == BlockType.TABLE)
    typ, tab = next(g)
    assert tab.name == "tab_ok"
    assert tab.df["a1"][1] == 3.14
    assert tab.df["a2"][1] == pd.to_datetime("2020-08-12")

    with pytest.raises(ValueError):
        typ, tab = next(g)
        print(f"-oOo-: {typ} {tab}")


def test_converter():
    """ Unit test
        Verify that misc. float input are handled consistently
    """
    # fmt: off
    table_lines = [
        ["**flt_errors"],
        ["dst1"],
        [ "a1"  , "a2"  , "a3"  , "a4"  ],
        [ "-"   , "-"   , "-"   , "-"   ],
        [ "NaN" , "nan" , "Nine", "Ten" ],
        [ 1     , 2     , 3     , 3.14  ],
    ]
    # fmt: on
    cf = custom_test_fixer()
    pandas_pdtab = make_table(table_lines, fixer=cf)
    js_obj = table_to_json_data(pandas_pdtab)
    assert js_obj["columns"]["a3"]["values"][0] is None
    assert js_obj["columns"]["a4"]["values"][1] == 3.14

    assert cf.fixes == 2  # Nine and Ten



import io
from pathlib import Path
from textwrap import dedent

from pytest import fixture

from pdtable import Table, TableDataFrame
from pdtable import TableBundle, read_csv, read_excel
from pdtable.demo.unit_converter import convert_this
from pdtable.utils import read_bundle_from_csv


def input_dir() -> Path:
    return Path(__file__).parent / "input"


@fixture
def csv_data():
    # fmt off
    return dedent(
        """\
        **farm_types1;;;
        your_farm my_farm farms_galore;;;
        species;  num;  flt;    log;
        text;       -;   kg;  onoff;
        chicken;    2;    3;      1;
        pig;        4;   39;      0;
        goat;       4;    -;      1;
        zybra;      4;    -;      0;
        cow;      NaN;  200;      1;
        goose;      2;    9;      0;

        **unrelated_table;;;
        your_farm my_farm farms_galore;;;
        species;  num;  flt;    log;
        text;       -;   kg;  onoff;
        unicorn;    2;    3;      1;
        """
    )
    # fmt on


def test_read_bundle_from_csv(csv_data):
    bundle = read_bundle_from_csv(io.StringIO(csv_data))
    # Correct number of tables read
    assert len(bundle) == 2
    # Correct values read
    assert bundle["farm_types1"]["flt"].unit == "kg"
    assert bundle["farm_types1"]["flt"].values[4] == 200
    assert bundle["unrelated_table"]["flt"].unit == "kg"
    assert bundle["unrelated_table"]["flt"].values[0] == 3


def test_read_bundle_from_csv__converts_units(csv_data):
    unit_dispatcher = {"farm_types1": {"flt": "g"}}
    bundle = read_bundle_from_csv(
        io.StringIO(csv_data), convert_units_to=unit_dispatcher, unit_converter=convert_this
    )
    # Units converted where dispatched
    assert bundle["farm_types1"]["flt"].unit == "g"
    assert bundle["farm_types1"]["flt"].values[4] == 200000
    # Units not converted where not dispatched
    assert bundle["unrelated_table"]["flt"].unit == "kg"
    assert bundle["unrelated_table"]["flt"].values[0] == 3


def test_TableBundle_from_file():
    """ Verify that TableBundle can be generated from top level API methods: read_csv, read_excel
    """
    input_file = input_dir() / "bundle.csv"
    bundle = TableBundle(read_csv(input_file), as_dataframe=True)
    assert bundle is not None
    assert len(bundle) == 3
    assert isinstance(bundle[0], TableDataFrame)

    assert bundle.unique("spelling_numbers").spelling[1] == "six"
    assert bundle[1].spelling[0] == "one"
    assert len(bundle.all("places_to_go")) == 2

    bundle = TableBundle(read_csv(input_file), as_dataframe=False)
    assert bundle is not None
    assert len(bundle) == 3
    assert isinstance(bundle[1], Table)
    assert bundle.spelling_numbers["spelling"].values[0] == "one"
    assert len(bundle.all("places_to_go")) == 2

    input_file = input_dir() / "bundle.xlsx"
    bundle = TableBundle(read_excel(input_file), as_dataframe=False)
    assert bundle is not None
    assert len(bundle) == 3
    assert isinstance(bundle[1], Table)
    assert bundle.spelling_numbers["spelling"].values[0] == "one"
    assert len(bundle.all("places_to_go")) == 2

    bundle = TableBundle(read_excel(input_file), as_dataframe=True)
    assert bundle is not None
    assert len(bundle) == 3
    assert isinstance(bundle[0], TableDataFrame)

    assert bundle.unique("spelling_numbers").spelling[1] == "six"
    assert bundle[1].spelling[0] == "one"
    assert len(bundle.all("places_to_go")) == 2



"""Machinery to parse columns in accordance with their unit indicator.

Parsers to convert column values of uncontrolled data types into values with a data type
consistent with the intended representation given the column's unit indicator.

A data-type-specific parser is implemented for each of the allowable StarTable column unit
indicators:
- 'text' -> str
- 'onoff' -> bool
- 'datetime' -> datetime / NaT
- everything else -> float / NaN

The parse_column() wrapper takes care of switching between these data-type-specific parsers.
This wrapper is the intended API.
"""
import datetime
from collections import defaultdict
from typing import Iterable, Sequence

import numpy as np
import pandas as pd

from .fixer import ParseFixer


def normalize_if_str(val):
    """If it's a string, strip it and make it lowercase. Otherwise, leave it alone."""
    return val.strip().lower() if isinstance(val, str) else val


def is_missing_data_marker(normalized_val):
    """Return True if, after normalization, it's a valid StarTable missing-data marker"""
    return normalize_if_str(normalized_val) in {"-", "nan"}


def _parse_text_column(values: Iterable, fixer: ParseFixer = None):
    # Ensure that 'values' is a Sequence, else np.array() will not unpack it
    return np.array(values if isinstance(values, Sequence) else list(values), dtype=np.str)


def _onoff_to_bool(val) -> bool:
    """Converts typical onoff columns values to bools"""
    conversions = {
        0: False,
        1: True,
        False: False,
        True: True,
        "0": False,
        "1": True,
        "false": False,
        "true": True,
    }
    return conversions[normalize_if_str(val)]


def _parse_onoff_column(values: Iterable, fixer: ParseFixer = None):
    bool_values = []
    for row, val in enumerate(values):
        try:
            bool_values.append(_onoff_to_bool(val))
        except KeyError as err:
            if fixer is not None:
                fixer.table_row = row
                fix_value = fixer.fix_illegal_cell_value("onoff", val)
                bool_values.append(fix_value)
            else:
                raise ValueError("Illegal value in onoff column", val) from err
    return np.array(bool_values, dtype=np.bool)


def _float_convert(val: str) -> float:
    if val in {"nan", "-"}:
        return np.nan
    return float(val)


def _parse_float_column(values: Iterable, fixer: ParseFixer = None):
    float_values = []
    for row, val in enumerate(values):
        if isinstance(val, float) or isinstance(val, int):
            # It's already a number.
            float_values.append(float(val))
            continue

        # It's something else than a number.
        if isinstance(val, str):
            # It's a string.
            val = normalize_if_str(val)
            try:
                # Parsing the string as one of the expected things (a number or missing value)
                float_values.append(_float_convert(val))
            except (KeyError, ValueError) as err:
                if fixer is not None:
                    fixer.table_row = row
                    fix_value = fixer.fix_illegal_cell_value("float", val)
                    float_values.append(fix_value)
                else:
                    raise ValueError("Illegal value in numerical column", val) from err
        else:
            # It isn't even a string. WTF let the fixer have a shot at it.
            if fixer is not None:
                fixer.table_row = row
                fix_value = fixer.fix_illegal_cell_value("float", val)
                float_values.append(fix_value)
            else:
                raise ValueError("Illegal value in numerical column", val)

    return np.array(float_values)


def _to_datetime(val):
    return pd.NaT if val in ["-", "nan"] else pd.to_datetime(val, dayfirst=True)


def _parse_datetime_column(values: Iterable, fixer: ParseFixer = None):
    datetime_values = []
    for row, val in enumerate(values):
        if isinstance(val, datetime.datetime):
            # It's already a datetime
            datetime_values.append(val)
            continue

        # It's something else than a datetime. Presumably a string.
        # TODO Fails when val is None. There is ambiguity about this None: if it came from a JsonData, it could be a legal NaT, but from Excel it could be an illegal empty cell!  # noqa:E501
        if val is None:
            # fixer should always be defined (= default ParseFixer)
            # when used via our top level API (read_csv, parse_blocks &c.)
            if fixer is not None:
                fixer.table_row = row
                fix_value = fixer.fix_illegal_cell_value("datetime", val)
                datetime_values.append(fix_value)
                continue
            else:
                raise ValueError(f"Illegal value in datetime column {val}")

        if not isinstance(val, str):
            raise ValueError(f"Illegal value in datetime column {val}")

        val = val.strip()
        if len(val) > 0 and (val[0].isdigit() or val in ["-", "nan"]):
            try:
                # Parsing the string as one of the expected things (a datetime or missing value)
                datetime_values.append(_to_datetime(val))
            except ValueError as err:
                if fixer is not None:
                    fixer.table_row = row
                    fix_value = fixer.fix_illegal_cell_value("datetime", val)
                    datetime_values.append(fix_value)
                    continue
                else:
                    raise ValueError("Illegal value in datetime column", val) from err
        else:
            if fixer is not None:
                fixer.table_row = row
                fix_value = fixer.fix_illegal_cell_value("datetime", val)
                datetime_values.append(fix_value)
            else:
                raise ValueError("Illegal value in datetime column", val)

    return np.array(datetime_values)


_column_parsers = defaultdict(lambda: _parse_float_column)
_column_parsers["text"] = _parse_text_column
_column_parsers["onoff"] = _parse_onoff_column
_column_parsers["datetime"] = _parse_datetime_column


def parse_column(unit_indicator: str, values: Iterable, fixer: ParseFixer = None) -> np.ndarray:
    """Parses column values to the intended data type as per the column's unit indicator.

    Parses column values to a consistent internal representation.
    The parser is chosen by switching on unit_indicator.
    The parsed values are returned as a numpy array with a suitable dtype.

    Args:
        unit_indicator:
            The column's unit indicator, e.g. 'text', 'onoff', 'datetime', '-', 'kg'...

        values:
            Iterable of the column's values

        fixer:
            Optional

    Returns:
        Parsed values, placed in a numpy array of a suitable dtype.

    """
    return _column_parsers[unit_indicator](values, fixer)



import sys
from typing import List, Any
import numpy as np
import pandas as pd


class ParseFixer:
    """ base class for auto-correcting errors and irregularities when parsing StarTable data

        * Possible specialization:

        class fixErrors(ParseFixer):
            # augment existing method
            def fix_illegal_cell_value(self, vtype, value):
                db_store(self.TableName,(self.TableColumn,self.TableRow))
                dfval = ParseFixer.fix_illegal_cell_value(self, vtype, value)
                return dfval

    """

    # Store legend of what's fixed + API
    def __init__(self):
        self._dbg = False
        self._errors = 0
        self._warnings = 0
        self._stop_on_errors = 1

        # Context info
        self.origin = None
        self.table_name = None
        self.column_name = None
        self.table_row = None

    @property
    def verbose(self):
        """
        if verbose: print debug info in fix_* methods
        """
        return self._dbg

    @verbose.setter
    def verbose(self, value: bool):
        self._dbg = value

    @property
    def stop_on_errors(self):
        """
        Raise exception if errors detected in input
        """
        return self._stop_on_errors

    @stop_on_errors.setter
    def stop_on_errors(self, value: bool):
        if value:
            self._dbg = True
        self._stop_on_errors = value

    @property
    def fixes(self):
        """ Number of warnings and errors fixed in input """
        return self._errors + self._warnings

    def reset_fixes(self):
        """ reset warning and error count """
        self._errors = 0
        self._warnings = 0

    def fix_duplicate_column_name(self, column_name: str, input_columns: List[str]) -> str:
        """
            The column_name already exists in  input_columns
            This method should provide a unique replacement name

        """
        if self.verbose:
            print(
                f"ParseFixer: fix duplicate column ({self.column_name}) {column_name} "
                f"in table: {self.table_name}"
            )

        self._errors += 1
        for sq in range(1000):
            test = f"{column_name}_fixed_{sq:03}"
            if test not in input_columns:
                return test

        return "{column_name}-fixed"

    def fix_missing_column_name(self, input_columns: List[str]) -> str:
        """
            The column_name: self.TableColumn is empty
            This method should provide a unique replacement name
        """
        if self.verbose:
            print(
                f"ParseFixer: fix missing column ({self.column_name}) {input_columns} "
                f"in table: {self.table_name}"
            )
        return self.fix_duplicate_column_name("missing", input_columns)

    def fix_missing_rows_in_column_data(
        self, row: int, row_data: List[str], num_columns: int
    ) -> List[str]:
        """
            The row is expected to have num_columns values
            This method should return the entire row of length num_columns
            by providing the missing default values
        """
        if self.verbose:
            print(f"ParseFixer: fix missing data in row ({row}) in table: {self.table_name}")
        row_data.extend(["NaN" for cc in range(num_columns - len(row_data))])
        self._errors += 1
        return row_data

    def fix_illegal_cell_value(self, vtype: str, value: str) -> Any:
        """
            The string value can not be converted to type vtype
            This method should return a suitable default value of type vtype

            Supported vtypes in { "onoff", "datetime", "-", "float" }
        """
        # TODO value can be something else than a string if it comes from e.g. Excel/openpyxl
        # TODO should not try to fix things that are illegal by design e.g. illegal empty cells
        defaults = {"onoff": False, "datetime": pd.NaT, "float": np.NaN, "-": np.NaN}
        if self.verbose:
            print(f'ParseFixer: illegal {vtype} value "{value}" in table {self.table_name}')
        default_val = defaults.get(vtype)
        self._warnings += 1
        if default_val is not None:
            return default_val
        else:
            return defaults["-"]

    def report(self):
        """ Inform user on stdout, stderr of any warnings / errors
        """
        if self.fixes > 0 and self.stop_on_errors:
            txt = f"Error(s): stop after {self.fixes} errors in input table '{self.table_name}'"
            raise ValueError(txt)

        if hasattr(self, "_called_from_test"):
            # TODO intended behaviour when _called_from_test = something else than True?
            return

        if self._warnings > 0:
            print(
                f"\nWarning: {self._warnings} data errors fixed while parsing "
                f"table '{self.table_name}'\n"
            )

        if self._errors > 0:
            sys.stderr.write(
                f"\nError: {self._errors} column errors fixed while parsing "
                f"table '{self.table_name}'\n"
            )



import pandas as pd

from typing import Iterable


def _represent_row_elements(row: Iterable, units: Iterable, na_rep: str = "-"):
    """Prepares row element representations for writing.

    In preparation for writing, coerce row values to representations compliant with
    the StarTable standard, in accordance with the values' respective column units:

    - NaN-like things in non-text columns are coerced to the specified nan representation
    - 'onoff' column values are coerced to 0's and 1's (where possible)
    - 'text' column values are coerced to strings
    - If the first column is 'text', its empty strings are replaced with an arbitrary but
      reasonable sealant

    Values are not, in general, converted to strings. If writing to a string format,
    stringification must be done by the client code.
    """
    for col, (val, unit) in enumerate(zip(row, units)):
        if unit != "text" and pd.isna(val):
            # Represent NaN-like things, except leave them be in text columns
            yield na_rep
        elif unit == "onoff":
            # Represent obvious booleans as 0's and 1's
            if val in [True, 1]:
                yield 1
            elif val in [False, 0]:
                yield 0
            else:
                # If it isn't an obvious boolean, leave it be
                yield val
        elif unit == "text":
            if val == "" and col == 0:
                # Prevent illegal empty string in first column
                yield "-"  # some arbitrary but reasonable sealant
            else:
                # Coerce everything to strings
                yield str(val)
        else:
            # Leave everything else be as it is
            yield val



"""Demo unit conversion"""
from typing import Tuple, Union, Iterable

import numpy as np


def convert_this(
    value: Union[float, np.ndarray], from_unit: str, to_unit: str = "...I guess you want base units"
) -> Tuple[Union[float, np.ndarray], str]:
    """
    A simple unit converter that hasn't read a lot of books.

    This is an example of how the user could implement a unit converting function.
    Here we define a few unit conversions manually. But the user could also choose to just
    delegate this to some dedicated unit support package like 'pint' or 'unum' or whatever.
    As long as everything is wrapped in a function with the same signature as this one.

    Args:
        value:
            A number or array of numbers. Arrays must support math operators and functions
            operating on it element-wise.
        from_unit:
            Old unit
        to_unit:
            New unit to which to convert

    Returns:
        Number or array converted from old to new unit.

    """
    if to_unit == from_unit:
        # Null conversion.
        return value, to_unit

    # Here are the base units of the non-base units that I know
    base_units = {"mm": "m", "C": "K", "g": "kg"}
    # Moreover, base units are, of course, their own base units
    base_units.update({bu: bu for bu in base_units.values()})

    # Here are a few aliases, for support of British, American, and, not least, Canadian English
    # and why not Canadian French while we're at it.
    unit_aliases = {"meter": "m", "metre": "m", "mètre": "m"}
    from_unit = unit_aliases.get(from_unit, from_unit)
    to_unit = unit_aliases.get(to_unit, to_unit)

    if to_unit == "...I guess you want base units":
        if from_unit in base_units:
            to_unit = base_units[from_unit]
        else:
            raise KeyError(f"No base unit defined for this unit.", from_unit)

    requested_conversion = (from_unit, to_unit)
    available_conversions = {
        ("m", "mm"): lambda x: x * 1000,
        ("mm", "m"): lambda x: x / 1000,
        ("C", "K"): lambda x: x + 273.15,
        ("K", "C"): lambda x: x - 273.15,
        ("kg", "g"): lambda x: x * 1000,
        ("g", "kg"): lambda x: x / 1000,
    }
    if requested_conversion not in available_conversions:
        raise KeyError(f"I don't know how to convert from '{from_unit}' to '{to_unit}'")

    return available_conversions[requested_conversion](value), to_unit



from dataclasses import dataclass, field
from typing import Set, List, Optional, Dict, Union

import numpy
import pandas as pd


class InvalidNamingError(Exception):
    pass


class TableOrigin:
    """
    A TableOrigin instance uniquely defines the source of a Table instance.

    Subclasses should take care to define __str__.
    If possible, as_html() should be defined to include backlink to original input.
    """

    def as_html(self) -> str:
        return str(self)


class TableOriginCSV(TableOrigin):
    def __init__(self, file_name: str = "", row: int = 0):
        self._file_name = file_name
        self._row = row

    def __str__(self) -> str:
        return f'"{self._file_name}" row {self._row}'

    def __repr__(self) -> str:
        return f"TableOriginCSV({self})"


@dataclass
class TableMetadata:
    """
    Node in tree describing table sources.

    operation: Describes operation to create table, e.g. 'Created',
    'Loaded', 'Concatenated', 'Merged'

    Only parents or origin should be defined. Neither needs to be.
    """

    name: str
    destinations: Set[str] = field(default_factory=lambda: {"all"})
    operation: str = "Created"
    parents: List["TableMetadata"] = field(default_factory=list)
    origin: Optional[
        str
    ] = ""  # Should be replaced with a TableOrigin object to allow file-edit access

    def __str__(self):
        dst = (
            " for {{{}}}".format(", ".join(d for d in self.destinations))
            if self.destinations
            else ""
        )
        src = ""
        if self.origin:
            src = f" from {self.origin}"
        if self.parents:
            src = " from {{{}}}".format(",".join(f"\n{c}" for c in self.parents))
        return f'Table "{self.name}" {dst}. {self.operation}{src}'


class ColumnFormat:
    def __init__(self, specifier: Union[str, int]):
        """Specifies how to format the values in this column as strings.
        Args:
            specifier:
                Can be either of:
                * An int indicating the precision i.e. number of decimal places
                  e.g. 2 will format 123.456 as '123.46' and 42 as '42.00'
                * A standard format specifier conforming to the format specification mini-language:
                  https://docs.python.org/3/library/string.html#format-specification-mini-language
                  e.g. '14.3e' will format 123.456 as '     1.235e+02'
        """
        self.specifier = f".{specifier}f" if isinstance(specifier, int) else specifier

    def __str__(self):
        return self.specifier

    def __repr__(self):
        return f"{self.__class__.__name__}: '{self.specifier}'"


# See https://docs.scipy.org/doc/numpy/reference/generated/numpy.dtype.html
_unit_from_dtype_kind = {
    "b": "onoff",
    "i": "-",
    "u": "-",
    "f": "-",
    "M": "-",
    "O": "text",
    "S": "text",
    "U": "text",
}
_units_special = {"text", "onoff"}


def unit_from_dtype(dtype: numpy.dtype) -> str:
    try:
        return _unit_from_dtype_kind[dtype.kind]
    except KeyError:
        raise ValueError(
            "The numpy data type {dtype} is of kind {dtype.kind} which "
            "cannot be assigned a StarTable unit"
        )


@dataclass
class ColumnMetadata:
    """
    Column metadata is always stored in dic with name as key
    """

    unit: str
    display_unit: Optional[str] = None
    display_format: Optional[ColumnFormat] = None

    def check_dtype(self, dtype, context: Optional[str] = None):
        base_unit = unit_from_dtype(dtype)
        context_text = " in " + context if context else ""
        if base_unit in _units_special:
            if not base_unit == self.unit:
                raise Exception(
                    f"Column unit {self.unit} not equal to {base_unit} expected "
                    f"from data type {dtype}{context_text}"
                )
        elif self.unit in _units_special:
            raise Exception(
                f"Special column unit {self.unit} not applicable for "
                f"data type {dtype}{context_text}"
            )

    @classmethod
    def from_dtype(cls, dtype: numpy.dtype, **kwargs) -> "ColumnMetadata":
        """
        Will set column unit to '-', 'onoff', or 'text' depending on dtype
        """
        return cls(unit_from_dtype(dtype), **kwargs)

    def update_from(self, b: "ColumnMetadata"):
        self.unit = b.unit
        if not self.display_unit:
            self.display_unit = b.display_unit
        if not self.display_format and b.display_format:
            self.display_format = b.display_format.copy()

    def copy(self) -> "ColumnMetadata":
        c = ColumnMetadata(self.unit)
        c.update_from(self)
        return c


class ComplementaryTableInfo:
    """A ComplementaryTableInfo object is responsible for storing any table information
    not stored by native dataframe

    A TableDataFrame object is a dataframe with such a ComplementaryTableInfo object
    attached as metadata.
    """

    def __init__(
        self, table_metadata: TableMetadata, columns: Optional[Dict[str, ColumnMetadata]] = None
    ):
        self.metadata: TableMetadata = table_metadata
        self.columns: Dict[str, ColumnMetadata] = columns if columns is not None else {}
        # self.template # TODO Table template data should/could be included here
        # self.parametrization = None: Do not include, see discussion in module docs
        self._last_dataframe_state = None

    def __str__(self):
        return str(self.metadata)

    def _update_columns(self, df):
        columns = self.columns
        df_columns = df.columns.values
        df_cname_set = set(df_columns)

        # check for duplicate names in dataframe
        if not len(df_columns) == len(df_cname_set):
            raise InvalidNamingError("Duplicate column names not allowed for Table")

        # remove columns not in dataframe
        for name in set(columns.keys()) - df_cname_set:
            del columns[name]

        # update metadata
        for name in df_columns:
            dtype = df[name].dtype
            if name in columns:
                columns[name].check_dtype(dtype)
            else:
                columns[name] = ColumnMetadata.from_dtype(dtype)

    def _check_dataframe(self, df: pd.DataFrame):
        """
        Check that column register matches columns of dataframe
        """
        dataframe_state = df.dtypes
        if dataframe_state.equals(self._last_dataframe_state):
            return
        self._update_columns(df)
        self._last_dataframe_state = dataframe_state



# TODO Get rid of this "utils" module; find a better home for its constituents.
# TODO These functions seem out of sync with the rest of the pdtable.io API. Needs revisiting.
from os import PathLike
from typing import Optional, Any, Tuple, Iterable, Dict, Callable, Union, TextIO

from pdtable import TableBundle, BlockType, Table
from pdtable import read_csv
from pdtable.proxy import UnitConverter, ColumnUnitDispatcher

# type hint shorthand
TableUnitDispatcher = Union[Dict[str, ColumnUnitDispatcher], Callable[[str], ColumnUnitDispatcher]]


def normalized_table_generator(
    block_gen: Iterable[Tuple[BlockType, Optional[Any]]],
    convert_units_to: TableUnitDispatcher = None,
    unit_converter: UnitConverter = None,
):
    for block_type, block in block_gen:
        if block is not None and block_type == BlockType.TABLE:
            # convert table units
            table: Table = block
            if isinstance(convert_units_to, Dict):
                to_units: ColumnUnitDispatcher = convert_units_to.get(table.name)
            elif isinstance(convert_units_to, Callable):
                to_units: ColumnUnitDispatcher = convert_units_to(table.name)
            else:
                raise TypeError(
                    "Table unit dispatcher of unexpected type.",
                    type(convert_units_to),
                    convert_units_to,
                )

            if to_units is not None:
                table = table.convert_units(to=to_units, converter=unit_converter)
            yield block_type, table

        else:
            yield block_type, block


def read_bundle_from_csv(
    input_path: Union[str, PathLike, TextIO],
    sep: Optional[str] = ";",
    convert_units_to: TableUnitDispatcher = None,
    unit_converter: UnitConverter = None,
) -> TableBundle:
    """Read single csv-file to TableBundle

    With a demo of bulk unit conversion of all tables at read time.

    unit_converter must accept units of type returned by the TableUnitDispatcher.
    """
    if convert_units_to and not unit_converter:
        raise ValueError("No unit converter supplied.")

    inputs = read_csv(input_path, sep)

    if convert_units_to is not None:
        inputs = normalized_table_generator(inputs, convert_units_to, unit_converter)

    return TableBundle(inputs)



from dataclasses import dataclass
from typing import List, Optional

import pdtable


class MetadataBlock(dict):
    def __init__(self, origin: Optional[str] = None):
        super().__init__()
        self.origin = origin

    def __repr__(self):
        sep = pdtable.CSV_SEP
        return "\n".join(f"{k}:{sep}{self[k]}{sep}" for k in self)


@dataclass
class Directive:

    name: str
    lines: List[str]
    origin: Optional[str] = None

    def __repr__(self):
        return f"***{self.name}{pdtable.CSV_SEP}\n" + "\n".join(self.lines)



import io
from textwrap import dedent

import pandas as pd

import pdtable
from pdtable import Table
from pdtable.io.csv import write_csv, _table_to_csv
from pdtable.table_metadata import ColumnFormat


def test__table_to_csv():
    # Make a table with content of various units
    t = Table(name="foo")
    t["place"] = ["home", "work", "beach", "wonderland"]
    t.add_column("distance", list(range(3)) + [float("nan")], "km")
    t.add_column(
        "ETA",
        pd.to_datetime(["2020-08-04 08:00", "2020-08-04 09:00", "2020-08-04 17:00", pd.NaT]),
        "datetime",
    )
    t.add_column("is_hot", [True, False, True, False], "onoff")

    # Write table to stream
    with io.StringIO() as out:
        _table_to_csv(t, out, ";", "-")
        # Assert stream content is as expected
        assert out.getvalue() == dedent(
            """\
            **foo;
            all
            place;distance;ETA;is_hot
            text;km;datetime;onoff
            home;0.0;2020-08-04 08:00:00;1
            work;1.0;2020-08-04 09:00:00;0
            beach;2.0;2020-08-04 17:00:00;1
            wonderland;-;-;0

            """
        )


def test_write_csv__writes_two_tables():
    # Make a couple of tables
    t = Table(name="foo")
    t["place"] = ["home", "work", "beach", "wonderland"]
    t.add_column("distance", list(range(3)) + [float("nan")], "km")
    t.add_column(
        "ETA",
        pd.to_datetime(["2020-08-04 08:00", "2020-08-04 09:00", "2020-08-04 17:00", pd.NaT]),
        "datetime",
    )
    t.add_column("is_hot", [True, False, True, False], "onoff")

    t2 = Table(name="bar")
    t2.add_column("number", [1, 6, 42], "-")
    t2.add_column("spelling", ["one", "six", "forty-two"], "text")

    # Write tables to stream
    with io.StringIO() as out:
        write_csv([t, t2], out)
        # Assert stream content is as expected
        assert out.getvalue() == dedent(
            """\
            **foo;
            all
            place;distance;ETA;is_hot
            text;km;datetime;onoff
            home;0.0;2020-08-04 08:00:00;1
            work;1.0;2020-08-04 09:00:00;0
            beach;2.0;2020-08-04 17:00:00;1
            wonderland;-;-;0

            **bar;
            all
            number;spelling
            -;text
            1;one
            6;six
            42;forty-two

            """
        )


def test_write_csv__leaves_stream_open_if_caller_passes_stream():
    # Make a table
    t2 = Table(
        pd.DataFrame({"number": [1, 6, 42], "spelling": ["one", "six", "forty-two"]}),
        name="bar",
        units=["-", "text"],
    )

    # Check write_csv single table and leave stream open for business
    with io.StringIO() as out:
        write_csv(t2, out)
        out.write("Fin\n")
        assert out.getvalue() == dedent(
            """\
            **bar;
            all
            number;spelling
            -;text
            1;one
            6;six
            42;forty-two

            Fin
            """
        )


def test_write_csv__writes_to_file(tmp_path):
    # Make a table
    t2 = Table(
        pd.DataFrame({"number": [1, 6, 42], "spelling": ["one", "six", "forty-two"]}),
        name="bar",
        units=["-", "text"],
    )

    # Write to file
    out_path = tmp_path / "write_csv_to_file.csv"
    write_csv(t2, out_path)

    # Now check file contents
    assert out_path.read_text() == dedent(
        """\
        **bar;
        all
        number;spelling
        -;text
        1;one
        6;six
        42;forty-two

        """
    )
    # Teardown
    out_path.unlink()


def test__write_csv__uses_altered_default_csv_separator(monkeypatch):

    # Change the default CSV separator
    # Using monkeypatch in lieu of just 'tables.CSV_SEP = ","'
    # so that this alteration of the default is only visible in this test.
    monkeypatch.setattr(pdtable, "CSV_SEP", ",")

    # Make a table with content of various units
    t = Table(
        pd.DataFrame(
            {
                "place": ["home", "work", "beach", "wonderland"],
                "distance": list(range(3)) + [float("nan")],
                "ETA": pd.to_datetime(
                    ["2020-08-04 08:00", "2020-08-04 09:00", "2020-08-04 17:00", pd.NaT]
                ),
                "is_hot": [True, False, True, False],
            }
        ),
        name="foo",
        units=["text", "km", "datetime", "onoff"],
    )

    # Write table to stream
    with io.StringIO() as out:
        write_csv(t, out)
        # Assert stream content is as expected
        assert out.getvalue() == dedent(
            """\
            **foo,
            all
            place,distance,ETA,is_hot
            text,km,datetime,onoff
            home,0.0,2020-08-04 08:00:00,1
            work,1.0,2020-08-04 09:00:00,0
            beach,2.0,2020-08-04 17:00:00,1
            wonderland,-,-,0

            """
        )


def test_write_csv__with_format_specs():
    # Make a table
    t2 = Table(
        pd.DataFrame(
            {"numbers": [1, 6, 42], "same_numbers": [1, 6, 42], "others": [1, 123.456, 42]}
        ),
        name="bar",
        units=["-", "-", "-"],
    )

    # Give formats to some columns, leave some without formats
    t2.column_metadata["same_numbers"].display_format = ColumnFormat(2)
    t2.column_metadata["others"].display_format = ColumnFormat("14.3e")

    with io.StringIO() as out:
        write_csv(t2, out)
        assert out.getvalue() == dedent(
            """\
            **bar;
            all
            numbers;same_numbers;others
            -;-;-
            1;1.00;     1.000e+00
            6;6.00;     1.235e+02
            42;42.00;     4.200e+01
            
            """
        )



from pathlib import Path
from textwrap import dedent

from ..auxiliary import Directive, MetadataBlock
from ..demo.directive_handlers import handle_includes
from ..io.parsers.blocks import parse_blocks
from ..store import BlockType


def test_metadata_block():
    ml = MetadataBlock()
    ml["author"] = "XYODA"
    ml["purpose"] = "Save the galaxy"
    assert repr(ml) == dedent(
        """\
        author:;XYODA;
        purpose:;Save the galaxy;"""
    )


def test_directive():
    d = Directive("foo", ["bar", "baz"])
    assert repr(d) == dedent(
        """\
        ***foo;
        bar
        baz"""
    )


def test_handle_includes():
    # TODO move directive handler stuff to a /demo package or something
    cell_rows = [
        line.split(";")
        for line in dedent(
            r"""
        ***include
        incl_1.csv
        incl_2.csv
    
        **t0
        all
        place;distance
        text;km
        home;0
        work;14
        beach;19
        """
        )
        .strip()
        .split("\n")
    ]

    bg = handle_includes(
        parse_blocks(cell_rows), input_dir=Path(__file__).parent / "input", recursive=True
    )
    bl = list(bg)
    tables = [b for t, b in bl if t == BlockType.TABLE]
    assert len(tables) == 4



import pandas as pd
import numpy as np

import pytest

from .. import Table, frame
from ..proxy import Column
from ..table_metadata import ColumnFormat


@pytest.fixture
def data_ab():
    return [{"cola": v, "colb": f"v{v}"} for v in range(4)]


@pytest.fixture
def data_cd():
    return [{"colc": v, "cold": f"d{v}"} for v in range(4)]


@pytest.fixture
def dft(data_ab):
    return frame.make_table_dataframe(data_ab, name="foo", destinations={"bar", "baz"})


def test_make_pdtable(data_ab):
    df = frame.make_table_dataframe(data_ab, name="foo")

    assert "cola" in df.columns
    assert df.cola[2] == 2

    data = frame.get_table_info(df)

    assert data.metadata.name == "foo"
    assert data.columns["cola"].unit == "-"
    assert data.columns["colb"].unit == "text"


def test_is_pdtable(dft, data_ab):
    df = pd.DataFrame(data_ab)
    assert not frame.is_table_dataframe(df)
    assert frame.is_table_dataframe(dft)


def test_get_table_data(dft):
    assert frame.get_table_info(dft).metadata.name == "foo"

    bad_table = frame.TableDataFrame()
    with pytest.raises(Exception):
        frame.get_table_info(bad_table)
    assert frame.get_table_info(bad_table, fail_if_missing=False) is None


def test_column(dft):
    c = Column(dft, "cola")
    assert c.unit == frame.get_table_info(dft).columns["cola"].unit
    c.unit = "m"
    assert c.unit == "m"
    assert c.unit == frame.get_table_info(dft).columns["cola"].unit

    # pandas docs say that indirect assignment is flaky
    # c.values[2] = 7
    # assert dft.cola[2] == 7


def test_add_column(dft):
    frame.add_column(dft, "colc", [f"c{v}" for v in range(20, 24)], "text")
    assert dft.colc[0] == "c20"
    assert frame.get_table_info(dft).columns["colc"].unit == "text"


def test_table_init__doesnt_crash():
    Table(pd.DataFrame({"c": [1, 2, 3], "d": [4, 5, 6]}), name="table2", units=["m", "kg"])


def test_table(dft):
    t = Table(dft)

    assert t.name == "foo"
    assert t.destinations == {"baz", "bar"}
    assert frame.is_table_dataframe(t.df)

    assert t["cola"].unit == "-"
    t["cola"].unit = "km"
    assert frame.get_table_info(t.df).columns["cola"].unit == "km"

    t["colc"] = range(20, 24)
    assert "colc" in t.column_names
    assert t["colc"].unit == "-"


def test_df_operations(data_ab, data_cd):
    t_ab = frame.make_table_dataframe(pd.DataFrame(data_ab), name="ab")
    t_cd = frame.make_table_dataframe(pd.DataFrame(data_cd), name="cd")

    _ = pd.concat([t_ab, t_cd], axis=0, sort=False)  # vertical concat
    r = pd.concat([t_ab, t_ab], axis=0, sort=False, ignore_index=True)  # vertical concat
    assert r.shape == (8, 2)

    t_ab2 = frame.make_table_dataframe(pd.DataFrame(data_ab), name="ab")
    Table(t_ab2)["cola"].unit = "m"

    with pytest.raises(frame.InvalidTableCombineError):
        # Fail on units for cola
        _ = pd.concat([t_ab, t_ab2])


def test_table_equals():
    t_ref = Table(
        pd.DataFrame({"c": [1, np.nan, 3], "d": [4, 5, 6]}), name="table2", units=["m", "kg"]
    )

    # True if...
    # identical
    assert t_ref.equals(
        Table(pd.DataFrame({"c": [1, np.nan, 3], "d": [4, 5, 6]}), name="table2", units=["m", "kg"])
    )
    # itself
    assert t_ref.equals(t_ref)
    # same numerical value but different data type (int vs. float)
    assert t_ref.equals(
        Table(
            pd.DataFrame({"c": [1, np.nan, 3], "d": [4.0, 5.0, 6.0]}),
            name="table2",
            units=["m", "kg"],
        )
    )

    # False if different...
    # name
    assert not t_ref.equals(
        Table(
            pd.DataFrame({"c": [1, np.nan, 3], "d": [4, 5, 6]}), name="Esmeralda", units=["m", "kg"]
        )
    )
    # destination
    assert not t_ref.equals(
        Table(
            pd.DataFrame({"c": [1, np.nan, 3], "d": [4, 5, 6]}),
            name="table2",
            units=["m", "kg"],
            destinations={"here", "there", "everywhere"},
        )
    )
    # unit
    assert not t_ref.equals(
        Table(
            pd.DataFrame({"c": [1, np.nan, 3], "d": [4, 5, 6]}),
            name="table2",
            units=["football_fields", "kg"],
        )
    )
    # column name
    assert not t_ref.equals(
        Table(
            pd.DataFrame({"level7_GHOUL": [1, np.nan, 3], "d": [4, 5, 6]}),
            name="table2",
            units=["m", "kg"],
        )
    )
    # data value
    assert not t_ref.equals(
        Table(
            pd.DataFrame({"c": [666, np.nan, 3], "d": [4, 5, 6]}), name="table2", units=["m", "kg"]
        )
    )
    # data value (ever so slightly)
    assert not t_ref.equals(
        Table(
            pd.DataFrame({"c": [1.00000000000001, np.nan, 3], "d": [4, 5, 6]}),
            name="table2",
            units=["m", "kg"],
        )
    )
    # thing entirely
    assert not t_ref.equals("a string")
    assert not t_ref.equals(42)
    assert not t_ref.equals(None)
    assert not t_ref.equals(pd.DataFrame({"c": [1, np.nan, 3], "d": [4, 5, 6]}))


def test_column_format():
    assert ColumnFormat(2).specifier == ".2f"
    assert ColumnFormat("14.2e").specifier == "14.2e"

    assert str(ColumnFormat(2)) == ".2f"
    assert repr(ColumnFormat(2)) == "ColumnFormat: '.2f'"



"""Conversion to JSON-serializable data structure.
"""
import datetime
from typing import Union, Dict, List

import numpy as np

from pdtable.table_metadata import TableOriginCSV

# Typing alias:
# JSON-like data structure of nested dicts ("objects"), lists ("arrays"), and JSON-native values
JsonData = Union[Dict[str, "JsonData"], List["JsonData"], str, float, int, bool, None]
# Typing alias: Same as JsonData, extended with a few non-JSON-native but readily JSONable types
JsonDataPrecursor = Union[
    Dict[str, "JsonDataPrecursor"],
    List["JsonDataPrecursor"],
    np.ndarray,
    str,
    float,
    int,
    bool,
    None,
    datetime.datetime,
    TableOriginCSV,
]

_json_encodable_value_maps = {
    dict: lambda obj: {kk: to_json_serializable(obj[kk]) for kk in obj.keys()},
    list: lambda obj: [to_json_serializable(kk) for kk in obj],
    float: lambda obj: obj if (not np.isnan(obj)) else None,
    int: lambda obj: obj,
    str: lambda obj: obj,
    bool: lambda obj: obj,
    type(None): lambda obj: obj,
}


def to_json_serializable(obj: JsonDataPrecursor) -> JsonData:
    """Converts object to a JSON-serializable data structure.

    Converts an object to a JSON-serializable hierarchical data structure of
    nested dicts ("objects"), lists ("arrays"), and values directly mappable to JSON-native types
    {float, int, str, bool} as well as None ("null").

    The types mentioned above are left as is.
    Also, additional element types are supported, mapping as follows:
    - numpy array -> list
    - values of type {datetime, table origin} -> string representation thereof
    """
    object_type = type(obj)
    if object_type in _json_encodable_value_maps:
        return _json_encodable_value_maps[object_type](obj)

    # Vanilla JSON encoder will choke on this value type.
    # Represent value as a JSON-encoder-friendly type.
    if isinstance(obj, np.ndarray):
        if f"{obj.dtype}" == "float64":
            return [val if (not np.isnan(val)) else None for val in obj.tolist()]
            # Note: would fail for obj.ndim > 1, but this is never the case here (columns are 1 dim)
        else:
            return [to_json_serializable(val) for val in obj.tolist()]

    if isinstance(obj, TableOriginCSV):
        return str(obj._file_name)

    if isinstance(obj, datetime.datetime):
        jval = str(obj)
        return jval if jval != "NaT" else None

    raise NotImplementedError(
        "Converting this type to a JSON-encodable type not yet implemented", type(obj)
    )



"""Test behaviour related to optional dependencies.

This module must be located completely outside the pdtable package. Otherwise it is difficult
to control what gets imported when and in which order, since pdtable/__init__.py gets executed
before the contents of everything else in the pdtable/ dir, including the tests in pdtable/test.
"""
from pathlib import Path
from textwrap import dedent
from unittest import mock
import sys

from pytest import raises


def test_openpyxl_absent(tmp_path):
    """Test behaviour when openpyxl is not installed

    pdtable should be fully functional, except for Excel I/O
    """
    # Mock openpyxl not being installed
    with mock.patch.dict(sys.modules, {'openpyxl': None}):
        # Can import pdtable
        import pdtable

        with raises(ImportError):
            # Fails on first use of read_excel
            # list() is used to trigger the otherwise lazy generator; else doesn't even import.
            list(pdtable.read_excel(""))

        with raises(ImportError):
            pdtable.write_excel([], tmp_path / "foo.xlsx")


def test_pint_absent():
    """Test behaviour when pint is not installed

    pdtable should be fully functional, except the courtesy pint converter
    """
    with mock.patch.dict(sys.modules, {'pint': None}):
        # Can import
        import pdtable
        with raises(ImportError):
            # Fails on first use of pint_converter
            p = pdtable.PintUnitConverter()
            # assert p(1, "m", "mm") == (1000, "mm")



import pandas as pd

from pdtable.io._represent import _represent_row_elements


def test__represent_row_elements():
    units = ["text", "km", "datetime", "onoff"]
    # Standard stuff
    assert list(
        _represent_row_elements(("foo", 123, pd.to_datetime("2020-08-04 08:00"), True), units)
    ) == ["foo", 123, pd.to_datetime("2020-08-04 08:00"), 1]

    # With NaN-like things
    assert list(_represent_row_elements(("foo", float("nan"), pd.NaT, 1), units)) == [
        "foo",
        "-",
        "-",
        1,
    ]

    # Specifying how NaN-like things should be represented
    assert list(
        _represent_row_elements(("foo", float("nan"), pd.NaT, False), units, na_rep="NaN")
    ) == ["foo", "NaN", "NaN", 0]

    # Empty strings: replace illegal in first column, leave others
    assert list(_represent_row_elements(("", "", float("nan")), ["text", "text", "text"])) == [
        "-",
        "",
        "nan",
    ]



"""
The store module implements data structures for storing collections of tables.

Plan is to implement processing of a stream of StarBlockType tokens
in a generic way that can be reused across readers and storage backends.
Examples include:
- attach template tokens to previous table
- attach file level metadata to subsequent tables
- unit normalization
- directive handling
"""

import re
from collections import defaultdict
from enum import Enum, auto
from typing import Iterable, Tuple, Any, Iterator, Optional, Union, List

from .frame import TableDataFrame
from .proxy import Table


class TableNameNotUniqueInBundleError(LookupError):
    """Raised when trying to access a table from a bundle by unique name, when in fact there
    are multiple tables with this same name in the bundle."""

    pass


class BlockType(Enum):
    """
    An enumeration of the tokens types that may be emitted by a reader.

    Design note
    Members of this enum are used to tag token type to avoid introspection.
    To aid reusable generation of metadata, it could be relevant to include
    synthetic block types FILE_BEGIN/END, SHEET_BEGIN/END.
    """

    # TODO class BlockType does not belong in this module, should be moved. Same for BlockGenerator.

    DIRECTIVE = auto()
    TABLE = auto()  # Interface: TableType
    TEMPLATE_ROW = auto()
    METADATA = auto()
    BLANK = auto()


BlockIterator = Iterable[Tuple[BlockType, Optional[Any]]]

TableType = Union[Table, TableDataFrame]


class TableBundle:
    """
    Simple table store with no regard for destinations.

    Upon creation from a block generator, ignores everything but Table blocks.

    Table blocks can be supplied as Table objects or any alternative data structure e.g. JsonData.
    If supplied as Table objects, the table blocks can be stored as TableDataFrame by setting
    as_Table=False.
    """

    def __init__(self, block_gen: BlockIterator, as_dataframe: bool = False):

        # Dict of lists of tables; each list contains all tables that have a certain name
        self._tables_named = defaultdict(list)
        # List of tables indexed by the the order in which they are appear in the block generator
        self._tables_in_order = []
        for block_type, block in block_gen:
            if block_type != BlockType.TABLE:
                continue
            table = block

            # Extract table name
            # Could be e.g. JsonData or other alternative data structure
            if hasattr(table, "name"):
                name = table.name
            elif isinstance(table, dict) and isinstance(table.get("name"), str):
                name = table.get("name")
            elif isinstance(table, list) and len(table) > 1:
                cell0 = table[0][0]
                if isinstance(cell0, str):
                    mm = re.search(r"^\s*\*\*(\S+)\s*", cell0)
                    if mm:
                        name = mm.group(1)
                    else:
                        raise NotImplementedError(
                            f"TableBundle: unable to extract table name from "
                            f"cell0 {cell0} in cellgrid-like table"
                        )
            else:
                raise NotImplementedError(
                    f"TableBundle: unable to extract table name from Table of type({type(table)})"
                )

            if as_dataframe and hasattr(table, "df"):
                self._tables_named[name].append(table.df)
                self._tables_in_order.append(table.df)
            else:
                self._tables_named[name].append(table)
                self._tables_in_order.append(table)

    def __getattr__(self, name: str) -> TableType:
        return self.unique(name)

    def __getitem__(self, idx: Union[str, int]) -> TableType:
        """Get table by numerical index or by name.

        Allows dual syntax: bundle[0] as well as bundle['table_name']
        """
        if isinstance(idx, str):
            return self.unique(idx)
        if isinstance(idx, int):
            return self._tables_in_order[idx]

        raise TypeError(f"getitem of type: {type(idx)}")

    def __iter__(self) -> Iterator[str]:
        """Iterator over tables"""
        return iter(self._tables_in_order)

    def __len__(self):
        """Total number of tables in this bundle"""
        return sum(len(self._tables_named[name]) for name in self._tables_named)

    def unique(self, name: str) -> TableType:
        """Returns table if there is exactly one table with this name, raises error otherwise."""
        lst = self._tables_named.get(name)
        if lst is None:
            # Ain't no table with this name.
            raise KeyError()
        if len(lst) > 1:
            # This name ain't unique, human.
            raise TableNameNotUniqueInBundleError()
        # Found exactly one table with this name. Here you go.
        return lst[0]

    def all(self, name: str) -> List[TableType]:
        """Returns all tables with this name."""
        lst = self._tables_named.get(name)
        return lst if lst is not None else []



import datetime

import pandas as pd
import openpyxl

try:
    from openpyxl.worksheet.worksheet import Worksheet as OpenpyxlWorksheet
except ImportError:
    # openpyxl < 2.6
    from openpyxl.worksheet import Worksheet as OpenpyxlWorksheet  # noqa: F401

from pdtable import Table
from pdtable.io.excel import write_excel
from pdtable.io._excel_openpyxl import _append_table_to_openpyxl_worksheet


def test__append_table_to_openpyxl_worksheet():
    # Make a table with content of various units
    t = Table(name="foo")
    t["place"] = ["home", "work", "beach", "wonderland"]
    t.add_column("distance", list(range(3)) + [float("nan")], "km")
    t.add_column(
        "ETA",
        pd.to_datetime(["2020-08-04 08:00", "2020-08-04 09:00", "2020-08-04 17:00", pd.NaT]),
        "datetime",
    )
    t.add_column("is_hot", [True, False, True, False], "onoff")

    wb = openpyxl.Workbook()
    ws = wb.active

    # Act
    _append_table_to_openpyxl_worksheet(t, ws)

    # Assert worksheet looks as expected:
    # table header by row
    assert ws["A1"].value == "**foo"
    assert ws["A2"].value == "all"
    assert [ws.cell(3, c).value for c in range(1, 5)] == ["place", "distance", "ETA", "is_hot"]
    assert [ws.cell(4, c).value for c in range(1, 5)] == ["text", "km", "datetime", "onoff"]

    # table data by column
    assert [ws.cell(r, 1).value for r in range(5, 9)] == ["home", "work", "beach", "wonderland"]
    assert [ws.cell(r, 2).value for r in range(5, 9)] == [0, 1, 2, "-"]
    assert [ws.cell(r, 3).value for r in range(5, 9)] == list(
        pd.to_datetime(["2020-08-04 08:00", "2020-08-04 09:00", "2020-08-04 17:00"])
    ) + ["-"]
    assert [ws.cell(r, 4).value for r in range(5, 9)] == [1, 0, 1, 0]


def test_write_excel(tmp_path):
    # Make a couple of tables
    t = Table(name="foo")
    t["place"] = ["home", "work", "beach", "wonderland"]
    t.add_column("distance", list(range(3)) + [float("nan")], "km")
    t.add_column(
        "ETA",
        pd.to_datetime(["2020-08-04 08:00", "2020-08-04 09:00", "2020-08-04 17:00", pd.NaT]),
        "datetime",
    )
    t.add_column("is_hot", [True, False, True, False], "onoff")

    t2 = Table(name="bar")
    t2.add_column("digit", [1, 6, 42], "-")
    t2.add_column("spelling", ["one", "six", "forty-two"], "text")

    # Write tables to workbook, save, and re-load
    out_path = tmp_path / "foo.xlsx"
    write_excel([t, t2], out_path)
    wb = openpyxl.load_workbook(out_path)
    ws = wb.active

    # Assert loaded worksheet looks as expected:
    # - table header by row
    assert ws["A1"].value == "**foo"
    assert ws["A2"].value == "all"
    assert [ws.cell(3, c).value for c in range(1, 5)] == ["place", "distance", "ETA", "is_hot"]
    assert [ws.cell(4, c).value for c in range(1, 5)] == ["text", "km", "datetime", "onoff"]

    # - table data by column
    assert [ws.cell(r, 1).value for r in range(5, 9)] == ["home", "work", "beach", "wonderland"]
    assert [ws.cell(r, 2).value for r in range(5, 9)] == [0, 1, 2, "-"]
    for r, d in zip(
        range(5, 8), pd.to_datetime(["2020-08-04 08:00", "2020-08-04 09:00", "2020-08-04 17:00"])
    ):
        # workaround openpyxl bug: https://foss.heptapod.net/openpyxl/openpyxl/-/issues/1493
        # openpyxl adds a spurious microsecond to some datetimes.
        assert abs(ws.cell(r, 3).value - d) <= datetime.timedelta(microseconds=1)
    assert ws.cell(8, 3).value == "-"
    assert [ws.cell(r, 4).value for r in range(5, 9)] == [1, 0, 1, 0]

    # Teardown
    out_path.unlink()



from .. import Table
from ._json import to_json_serializable, JsonData
from .parsers.blocks import make_table


def json_data_to_table(table_json_data: JsonData, **kwargs) -> Table:
    """  translate table-dictionary (JSON-like) to Table

    kwargs:
        fixer: ParseFixer instance or class
    """
    lines_json = [
        [f'**{table_json_data["name"]}'],
        [" ".join(table_json_data["destinations"])],
        [f"{cname}" for cname in table_json_data["columns"].keys()],
        [f"{col['unit']}" for col in table_json_data["columns"].values()],
    ]
    data = [col["values"] for col in table_json_data["columns"].values()]

    json_rows = list(map(list, zip(*data)))  # transposed columns
    lines_json.extend(json_rows)
    # note: this allows us to use ParseFixer !
    return make_table(lines_json, origin="JsonData", **kwargs)


def table_to_json_data(table: Table) -> JsonData:
    """  translate Table to table-dictionary (JSON-like)
    """

    table_data = {
        "name": table.name,
        "destinations": {dst: None for dst in table.metadata.destinations},
        "columns": {},
    }
    for idx, cname in enumerate(table.column_names):
        table_data["columns"][cname] = {
            "unit": table.units[idx],
            "values": list(table.df[cname]),
        }
    return to_json_serializable(table_data)



# -*- coding: utf-8 -*-
# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.5.2
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Demo of `pdtable`
#
# The `pdtable` package allows working with StarTable tables as pandas dataframes while consistently keeping track of StarTable-specific metadata such as table destinations and column units.
#
# This is implemented by providing two interfaces to the same backing object:
# - `TableDataFrame` is derived from `pandas.DataFrame`. It carries the data and is the interface of choice for interacting with the data using the pandas API, with all the convenience that this entails.  It also carries StarTable metadata, but in a less accessible way.
# - `Table` is a stateless façade for a backing `TableDataFrame` object. `Table` is the interface of choice for convenient access to StarTable-specific metadata such as column units, and table destinations and origin. Some data manipulation is also possible in the `Table` interface, though this functionality is limited.
#
# ![Table interfaces](../docs/diagrams/img/table_interfaces/table_interfaces.svg)
#
# This notebook provides a demo of how this works. See the `pdtable` docstring for a discussion of the implementation.
#
# ## Idea
#
# The central idea is that as much as possible of the table information is stored as a pandas dataframe,
# and that the remaining information is stored as a `ComplementaryTableInfo` object attached to the dataframe as registered metadata.
# Further, access to the full table datastructure is provided through a facade object (of class `Table`). `Table` objects
# have no state (except the underlying decorated dataframe) and are intended to be created when needed and discarded afterwards:
#
# ```
# tdf = make_table_dataframe(...)
# unit_height = Table(tdf).height.unit
# ```
#
# Advantages of this approach are that:
#
# 1. Code can be written for (and tested with) pandas dataframes and still operate on `TableDataFrame` objects.
#    This frees client code from necessarily being coupled to the startable project.
#    A first layer of client code can read and manipulate StarTable data, and then pass it on as a (`TableDataFrame`-flavoured) `pandas.DataFrame` to a further layer having no knowledge of `pdtable`.
# 2. The access methods of pandas `DataFrames`, `Series`, etc. are available for use by consumer code via the `TableDataFrame` interface. This both saves the work
#    of re-implementing similar access methods on a StarTable-specific object, and likely allows better performance and documentation.

# %%
# %load_ext autoreload
# %autoreload 2
from textwrap import dedent

import pandas as pd

# %%
# Workaround for this demo in notebook form:
# Notebooks don't have repo root in sys.path
# Better solutions are welcome...
import sys
from pathlib import Path

if Path.cwd().name == "examples":
    # Maybe we're in a notebook in the /examples folder
    repo_root_dir = str(Path.cwd().parent)
    if repo_root_dir not in sys.path:
        sys.path.append(repo_root_dir)

# %% [markdown]
# ## The `Table` aspect
#
# The table aspect presents the table with a focus on StarTable metadata.
# Data manipulation functions exist, and more are easily added -- but focus is on metadata.

# %%
from pdtable import frame
from pdtable.proxy import Table

# %%
t = Table(name="mytable")

# Add columns explicitly...
t.add_column("places", ["home", "work", "beach", "unicornland"], "text")

# ...or via item access
t["distance"] = [0, 1, 2, 42]

# Modify column metadata through column proxy objets:
t["distance"].unit = "km"  # (oops I had forgotten to set distance unit!)

# Table renders as annotated dataframe
t

# %%
t.column_names

# %%
# Each column has associated metadata object that can be manipulated:
t["places"]

# %%
t.units

# %%
str(t)

# %%
t.metadata

# %%
# Creating table from pandas dataframe:
t2 = Table(pd.DataFrame({"c": [1, 2, 3], "d": [4, 5, 6]}), name="table2", units=["m", "kg"])
t2

# %% [markdown]
# ## The `TableDataFrame` aspect
#
# Both the table contents and metadata displayed and manipulated throught the `Table`-class is stored as a `TableDataFrame` object, which is a normal pandas dataframe with two modifications:
#
# * It has a `_table_data` field registered as a metadata field, so that pandas will include it in copy operations, etc.
# * It will preserve column and table metadata for some pandas operations, and fall back to returning a vanilla `pandas.DataFrame` if this is not possible/implemented.
#

# %%
# After getting a reference to the TableDataFrame backing Table t, it is safe to delete t:
df = t.df
del t
# Just construct a new Table facade, and everything is back in place.
Table(df)

# %%
# Interacting with table metadata form  without the Table proxy is slightly verbose
df2 = frame.make_table_dataframe(
    pd.DataFrame({"c": [1, 2, 3], "d": [4, 5, 6]}), name="table2", units=["m", "kg"]
)

# %%
# Example: combining columns from multiple tables
df_combinded = pd.concat([df, df2], sort=False, axis=1)
Table(df_combinded)

# %% [markdown]
# ## Fundamental issues with the facade approach
#
# The key problem with not integrating more tightly with the dataframe datastructure is that some operations will not trigger `__finalize__` and thus have to be deciphered based on the dataframe and metadata states.
#
# One example of this is column names:

# %%
# Make a table and get its backing df
tab = Table(pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]}), name="foo", units=["m", "kg"])
df = tab.df
# Edit df column names
df.columns = ["a_new", "b_new"]

# This edit doesn't propagate to the hidden metadata!
col_names_before_access = list(df._table_data.columns.keys())
print(f"Metadata columns before update triggered by access:\n{col_names_before_access}\n")
assert col_names_before_access == ["a", "b"]  # Still the old column names

# ... until the facade is built and the metadata is accessed.
_ = str(Table(df))  # access it
col_names_after_access = list(df._table_data.columns.keys())
print(f"Metadata columns after update:\n{col_names_after_access}\n")
assert col_names_after_access == ["a_new", "b_new"]  # Now they've updated

# %% [markdown]
# ## I/O
# StarTable data can be read from, and written to, CSV files and Excel workbooks. In its simplest usage, the pdtable.io API can be illustrated as:
# ![pdtable.io simplified API](..\docs\diagrams\img\io_simple\io_simple.svg)
# We will return to a more detailed diagram also illustrating JSON support.


# %% [markdown]
# ### Reading
# The functions `read_excel()` and `read_csv()` do what it says on the tin.
# `read_csv()` can read from files as well as text streams.
#
# These generate tuples of (`block_type`, `block`) where
# * `block` is a StarTable block
# * `block_type` is a `BlockType` enum that indicates which type of StarTable block `block` is. Possible values: `DIRECTIVE`, `TABLE`, `TEMPLATE_ROW`, `METADATA`, and `BLANK`.

# %% [markdown]
# Let's make some CSV StarTable data, put it in a text stream, and read that stream. (Could just as well have been a CSV file.)

# %%
from io import StringIO
from pdtable import read_csv

csv_data = StringIO(
    dedent(
        """\
    author: ;XYODA     ;
    purpose:;Save the galaxy

    ***gunk
    grok
    jiggyjag

    **places;
    all
    place;distance;ETA;is_hot
    text;km;datetime;onoff
    home;0.0;2020-08-04 08:00:00;1
    work;1.0;2020-08-04 09:00:00;0
    beach;2.0;2020-08-04 17:00:00;1

    **farm_animals;;;
    your_farm my_farm other_farm;;;
    species;n_legs;avg_weight;
    text;-;kg;
    chicken;2;2;
    pig;4;89;
    cow;4;200;
    unicorn;4;NaN;
    """
    )
)

# Read the stream. Syntax is the same if reading CSV file.
# Reader function returns a generator; unroll it in a list for convenience.
block_list = list(read_csv(csv_data))
assert len(block_list) == 4


# %% [markdown]
# The reader generates tuples of `(BlockType, block)`. Note that blank/comment lines are read but not parsed.

# %%
for bt, b in block_list:
    print(bt, type(b))

# %% [markdown]
# Here's one of the tables.

# %%
t = block_list[2][1]
assert t.name == "places"
print(t)

# %% [markdown]
# We can pick the tables (and leave out blocks of other types) like this:

# %%
from pdtable import BlockType

tables = [b for bt, b in block_list if bt == BlockType.TABLE]
assert len(tables) == 2


# %% [markdown]
# ### Filtered reading
# Blocks can be filtered prior to parsing, by passing to a callable as `filter` argument to the reader functions. This can
# reduce reading time substantially when reading only a few tables from an otherwise large file
# or stream.
#
# The callable must accept two arguments: block type, and block name. Only those blocks for which the filter callable returns `True` are fully parsed. Other blocks
# are parsed only superficially i.e. only the block's top-left cell, which is just
# enough to recognize block type and name to pass to the filter callable, thus avoiding the much more
# expensive task of parsing the entire block, e.g. the values in all columns and rows of a large
# table.
#
# Let's design a filter that only accepts tables whose name contains the word `'animal'`.

# %%
def is_table_about_animals(block_type: BlockType, block_name: str) -> bool:
    return block_type == BlockType.TABLE and "animal" in block_name


# %% [markdown]
# Now let's see what happens when we use this filter when re-reading the same CSV text stream as before.

# %%
csv_data.seek(0)
block_list = list(read_csv(csv_data, filter=is_table_about_animals))

assert len(block_list) == 1
block_list[0][1]

# %% [markdown]
# Non-table blocks were ignored, and so were table blocks that weren't animal-related.

# %% [markdown]
# ### Writing
# Use the aptly named `write_csv()` and `write_excel()`. Note that `write_csv()` can write to files as well as to text streams.
#
# Let's write the tables we read earlier to a text stream.

# %%
from pdtable import write_csv

with StringIO() as s:
    write_csv(tables, s)

    print(s.getvalue())


# %% [markdown]
# ### JSON support
# StarTable data can be converted to and from a `JsonData` object i.e. a JSON-ready data structure of nested dicts ("objects"), lists ("arrays"), and JSON-native values. This `JsonData` can then be serialized and deserialized directly using the standard library's `json.dump()` and `json.load()`.
#
# A `JsonData` representation is currently only defined for table blocks.
#
# The pdtable.io API including JSON support can be illustrated as a more detailed version of the diagram shown earlier:
#
# ![pdtable.io API](..\docs\diagrams\img\io_detailed\io_detailed.svg)
#
# In addition, two successive intermediate data structures are visible in this diagram:
#
# - "Raw cell grid" is simply a sequence of sequence of values (usually a list of lists or list of tuples) interpreted as rows of cells, with cell contents as read in their raw form from CSV or Excel.
# - `JsonDataPrecursor` is essentially the same thing as `JsonData`, with the exception that it may contain values that are not directly consumable by `json.dump()`. These values must be converted further if the aim is a `JsonData`; for example Numpy arrays to lists, `datetime.datetime` to a string representation thereof, etc.
#
# Let's re-read the CSV stream we created earlier, but specifying the argument `to='jsondata'` such that tables should be parsed to JSON data.

# %%
csv_data.seek(0)
block_gen = read_csv(csv_data, to="jsondata")
json_ready_tables = [b for bt, b in block_gen if bt == BlockType.TABLE]

json_ready_tables[0]

# %% [markdown]
# This data structure is now easily serialized as JSON:

# %%
import json

table_json = json.dumps(json_ready_tables[0])
print(table_json)

# %% [markdown]
# There are also utilities to convert back and forth between `JsonData` and `Table`.

# %%
from pdtable.io import json_data_to_table, table_to_json_data

t = json_data_to_table(json_ready_tables[0])
t  # It's now a Table

# %%
table_to_json_data(t)  # Now it's back to JsonData

# %% [markdown]
#

# %% [markdown]
# ### Handling directives
# StarTable Directive blocks are intended to be interpreted and handled at read time, and then discarded. The client code (i.e. you) is responsible for doing this. Handling directives is done outside of the `pdtable` framework. This would typically done by a function that consumes a `BlockGenerator` (the output of the reader functions), processes the directive blocks encountered therein, and in turn yields a processed `BlockGenerator`, as in:
#
# ```
# def handle_some_directive(bg: BlockGenerator, *args, **kwargs) -> BlockGenerator:
#     ...
# ```
#
# Usage would then be:
# ```
# block_gen = handle_some_directive(read_csv('foo.csv'), args...)
# ```
#
# Let's imagine a directive named `'include'`, which the client application is meant to interpret as: include the contents of the listed StarTable CSV files along with the contents of the file you're reading right now. Such a directive could look like:
# ```
# ***include
# bar.csv
# baz.csv
# ```
# Note that there's nothing magical about the name "include"; it isn't StarTable syntax. This name, and how such a directive should be interpreted, is entirely defined by the application. We could just as easily imagine a `rename_tables` directive requiring certain table names to be amended upon reading.
#
# But let's stick to the "include" example for now. The application wants to interpret the `include` directive above as: read two additional files and throw all the read blocks together. Perhaps even recursively, i.e. similarly handle any `include` directives encountered in `bar.csv` and `baz.csv` and so on. A handler for this could be designed to be used as
#
# ```
# block_gen = handle_includes(read_csv('foo.csv'), input_dir='./some_dir/', recursive=True)
# ```
#
# and look like this:

# %%
import functools
from pdtable import BlockIterator, Directive


def handle_includes(bg: BlockIterator, input_dir, recursive: bool = False) -> BlockIterator:
    """Handles 'include' directives, optionally recursively.

    Handles 'include' directives.
    'include' directives must contain a list of files located in directory 'input_dir'.

    Optionally handles 'include' directives recursively. No check is done for circular references.
    For example, if file1.csv includes file2.csv, and file2.csv includes file1.csv, then infinite
    recursion ensues upon reading either file1.csv or file2.csv with 'recursive' set to True.

    Args:
        bg:
            A block generator returned by read_csv

        input_dir:
            Path of directory in which include files are located.

        recursive:
            Handle 'include' directives recursively, i.e. 'include' directives in files themselves
            read as a consequence of an 'include' directive, will be handled. Default is False.

    Yields:
        A block generator yielding blocks from...
        * if recursive, the entire tree of files in 'include' directives.
        * if not recursive, the top-level file and those files listed in its 'include' directive (if
          any).

    """

    deep_handler = (
        functools.partial(handle_includes, input_dir=input_dir, recursive=recursive)
        if recursive
        else lambda x: x
    )

    for block_type, block in bg:
        if block_type == BlockType.DIRECTIVE:
            directive: Directive = block
            if directive.name == "include":
                # Don't emit this directive block; handle it.
                for filename in directive.lines:
                    yield from deep_handler(read_csv(Path(input_dir) / filename))
            else:
                yield block_type, block
        else:
            yield block_type, block


# %% [markdown]
# This is only an example implementation. You're welcome to poach it, but note that no check is done for circular references! E.g. if `bar.csv` contains an `include` directive pointing to `foo.csv` then calling `handle_includes` with `recursive=True` will result in a stack overflow. You may want to perform such a check if relevant for your application.

# %%



from .csv import read_csv, write_csv  # noqa: F401
from .excel import read_excel, write_excel  # noqa: F401
from .json import table_to_json_data, json_data_to_table  # noqa: F401
from .parsers.fixer import ParseFixer  # noqa: F401
from .parsers.blocks import parse_blocks  # noqa: F401



import functools
from pathlib import Path

from ..auxiliary import Directive
from ..io.csv import read_csv
from ..store import BlockIterator, BlockType


def handle_includes(bg: BlockIterator, input_dir, recursive: bool = False) -> BlockIterator:
    """Handles 'include' directives, optionally recursively.

    Handles 'include' directives.
    'include' directives must contain a list of files located in directory 'input_dir'.

    Optionally handles 'include' directives recursively. No check is done for circular references.
    For example, if file1.csv includes file2.csv, and file2.csv includes file1.csv, then infinite
    recursion ensues upon reading either file1.csv or file2.csv with 'recursive' set to True.

    Args:
        bg:
            A block generator returned by read_file_csv

        input_dir:
            Path of directory in which include files are located.

        recursive:
            Handle 'include' directives recursively, i.e. 'include' directives in files themselves
            read as a consequence of an 'include' directive, will be handled. Default is False.

    Yields:
        A block generator yielding blocks from...
        * if recursive, the entire tree of files in 'include' directives.
        * if not recursive, the top-level file and those files listed in its 'include' directive (if
          any).

    """

    deep_handler = (
        functools.partial(handle_includes, input_dir=input_dir, recursive=recursive)
        if recursive
        else lambda x: x
    )

    for block_type, block in bg:
        if block_type == BlockType.DIRECTIVE:
            directive: Directive = block
            if directive.name == "include":
                for filename in directive.lines:
                    yield from deep_handler(read_csv(Path(input_dir) / filename))
            else:
                yield block_type, block
        else:
            yield block_type, block



import pytest

from pdtable import TableBundle, Table, TableDataFrame
from pdtable.io.parsers.blocks import parse_blocks

cell_rows = [
    # fmt off
    ["**foo"],
    ["all"],
    ["column"],
    ["text"],
    ["bar"],
    ["zoo"],
    [],
    ["::", "Table foo describes"],
    [None, "the fooness of things"],
    [":.column", "Column is a column in foo"],
    [],
    ["**infs"],
    ["all"],
    ["file_bytes", "file_date", "has_table"],
    ["-", "text", "onoff"],
    [15373, "a", 0],
    [15326, "b", 1],
    []
    # fmt on
]


def test_bundle_from_csv():

    bundle = TableBundle(parse_blocks(cell_rows), as_dataframe=True)

    assert bundle.foo.column.values[0] == "bar"


def test_TableBundle_as_dataframe():
    """ Verify that as_dataframe is functioning as expected (switch TableType)
    """

    # pdtable generator
    bundle = TableBundle(parse_blocks(cell_rows, to="pdtable"), as_dataframe=True)
    assert bundle.infs.file_bytes.values[1] == 15326.0
    assert bundle is not None
    assert len(bundle) == 2
    assert isinstance(bundle[0], TableDataFrame)

    # pdtable generator
    bundle = TableBundle(parse_blocks(cell_rows, to="pdtable"), as_dataframe=False)
    assert bundle.infs["file_bytes"].values[1] == 15326.0
    assert bundle is not None
    assert len(bundle) == 2
    assert isinstance(bundle[1], Table)

    # do not error on other table types
    bundle = TableBundle(parse_blocks(cell_rows, to="cellgrid"), as_dataframe=True)
    assert bundle is not None
    assert isinstance(bundle[0], list)  # cellgrid


def test_TableBundle_iterator():
    """ Verify that iterator is functioning as expected
    """
    bundle = TableBundle(parse_blocks(cell_rows, to="pdtable"))
    count = 0
    seen = {}
    for tab in bundle:
        assert type(tab) is Table
        seen[tab.name] = tab
        count += 1
    assert count == 2
    assert len(seen) == 2
    assert seen["foo"] is not None
    assert seen["infs"] is not None

    """ Verify that we can iterate other types than pdtable
    """
    bundle = TableBundle(parse_blocks(cell_rows, to="cellgrid"))
    count = 0
    for tab in bundle:
        assert type(tab) is list
        assert tab[0][0] in {"**foo", "**infs"}
        count += 1
    assert count == 2
    assert bundle["foo"] is not None
    assert bundle["infs"] is not None

    bundle = TableBundle(parse_blocks(cell_rows, to="jsondata"))
    count = 0
    for tab in bundle:
        assert type(tab) is dict
        assert tab["name"] in {"foo", "infs"}
        count += 1
    assert count == 2
    assert bundle["foo"] is not None
    assert bundle["infs"] is not None


def test_TableBundle_unique():
    """ Verify that unique() is functioning as expected
    """
    bundle1 = TableBundle(parse_blocks(cell_rows))
    # bundle1 now contains one 'foo' and one 'infs'
    assert len(bundle1) == 2

    with pytest.raises(LookupError):
        tab = bundle1.unique("-not there-")

    tab = bundle1.unique("foo")
    assert tab.name == "foo"

    tab = bundle1.unique("infs")
    assert tab.name == "infs"

    cells2 = []
    cells2.extend(cell_rows)
    cells2.extend([])
    cells2.extend(cell_rows)

    bundle2 = TableBundle(parse_blocks(cells2))
    # bundle2 now contains two 'foo' and two 'infs'
    assert len(bundle2) == 4

    with pytest.raises(LookupError):
        tab = bundle2.unique("-not there-")

    with pytest.raises(LookupError):
        tab = bundle2.unique("foo")

    with pytest.raises(LookupError):
        tab = bundle2.unique("infs")


def test_TableBundle_getitem():
    """ Verify that unique() is functioning as expected
    """
    bundle1 = TableBundle(parse_blocks(cell_rows))
    # bundle1 now contains one 'foo' and one 'infs'
    assert len(bundle1) == 2

    with pytest.raises(LookupError):
        tab = bundle1["-not there-"]

    # verify getitem
    with pytest.raises(TypeError):
        tab = bundle1[bundle1]

    # hashed
    tab = bundle1["foo"]
    assert tab.name == "foo"

    tab = bundle1["infs"]
    assert tab.name == "infs"

    # indexed
    tab = bundle1[0]
    assert tab.name == "foo"

    tab = bundle1[1]
    assert tab.name == "infs"

    with pytest.raises(IndexError):
        tab = bundle1[2]

    cells2 = []
    cells2.extend(cell_rows)
    cells2.extend([])
    cells2.extend(cell_rows)

    bundle2 = TableBundle(parse_blocks(cells2))
    # bundle2 now contains two 'foo' and two 'infs'
    assert len(bundle2) == 4

    with pytest.raises(LookupError):
        tab = bundle2["-not there-"]

    with pytest.raises(LookupError):
        tab = bundle2["foo"]

    with pytest.raises(LookupError):
        tab = bundle2["infs"]

    # indexed
    tab = bundle2[0]
    assert tab.name == "foo"

    tab = bundle2[1]
    assert tab.name == "infs"

    tab = bundle2[2]
    assert tab.name == "foo"

    tab = bundle2[3]
    assert tab.name == "infs"

    with pytest.raises(IndexError):
        tab = bundle2[4]


def test_TableBundle_all():
    """ Verify that all() is functioning as expected
    """
    bundle1 = TableBundle(parse_blocks(cell_rows))
    # bundle1 now contains one 'foo' and one 'infs'
    assert len(bundle1) == 2

    lst = bundle1.all("-not there-")
    assert len(lst) == 0

    lst = bundle1.all("foo")
    assert len(lst) == 1
    for tab in lst:
        assert tab.name == "foo"

    lst = bundle1.all("infs")
    assert len(lst) == 1
    for tab in lst:
        assert tab.name == "infs"

    cells2 = []
    cells2.extend(cell_rows)
    cells2.extend([])
    cells2.extend(cell_rows)

    bundle2 = TableBundle(parse_blocks(cells2))
    # bundle2 now contains two 'foo' and two 'infs'
    assert len(bundle2) == 4

    lst = bundle2.all("-not there-")
    assert len(lst) == 0

    lst = bundle2.all("foo")
    assert len(lst) == 2
    for tab in lst:
        assert tab.name == "foo"

    lst = bundle2.all("infs")
    assert len(lst) == 2
    for tab in lst:
        assert tab.name == "infs"



"""Interface to read/write Tables from/to an Excel workbook.

The only Excel I/O engine supported right now is 'openpyxl', but this module can
be extended to support others readers such as 'xlrd' and writers such as 'xlsxwriter'.

openpyxl (and eventually other engines) are not required at install time; only when the functions
requiring them (read_excel() or write_excel()) are called for the first time.

"""
import os
import io
from os import PathLike
from typing import Union, Callable, Iterable, BinaryIO

from .parsers.blocks import parse_blocks
from .parsers.fixer import ParseFixer
from .. import BlockType, Table, TableBundle
from ..store import BlockIterator


def read_excel(
    source: Union[str, PathLike, BinaryIO],
    origin=None,
    fixer: ParseFixer = None,
    to: str = "pdtable",
    filter: Callable[[BlockType, str], bool] = None,
) -> BlockIterator:
    """Reads StarTable blocks from an Excel workbook.
    # TODO copy most of read_csv() docstring over

    Reads StarTable blocks from an Excel workbook file at the specified path.
    Yields them one at a time as a tuple: (block type, block content)

    Args:
        source:
            Path of workbook to read.



    Yields:
        Tuples of the form (block type, block content)
    """

    kwargs = {"origin": origin, "fixer": fixer, "to": to, "filter": filter}

    try:
        from ._excel_openpyxl import read_cell_rows_openpyxl as read_cell_rows

    except ImportError as err:
        raise ImportError(
            "Unable to find a usable Excel engine. "
            "Tried using: 'openpyxl'.\n"
            "Please install openpyxl for Excel I/O support."
        ) from err

    if not isinstance(source, (str, PathLike)):
        assert isinstance(source, io.BufferedIOBase)

    yield from parse_blocks(read_cell_rows(source), **kwargs)


def write_excel(
    tables: Union[Table, Iterable[Table], TableBundle],
    path: Union[str, os.PathLike],
    na_rep: str = "-",
):
    """Writes one or more tables to an Excel workbook.

    Writes table blocks to an Excel workbook file.
    Values are formatted to comply with the StarTable standard where necessary and possible.

    This is a thin wrapper around parse_blocks(). The only thing it does is to present the contents
    of an Excel workbook as a Iterable of cell rows, where each row is a sequence of values.

    Args:
        tables:
            Table(s) to write. Can be a single Table or an iterable of Tables.
        path:
            File path to which to write.
        na_rep:
            Optional; String representation of missing values (NaN, None, NaT).
            If overriding the default '-', it is recommended to use another value compliant with
            the StarTable standard.
    """
    try:
        from ._excel_openpyxl import write_excel_openpyxl as write_excel_func

    except ImportError as err:
        raise ImportError(
            "Unable to find a usable spreadsheet engine. "
            "Tried using: 'openpyxl'.\n"
            "Please install openpyxl for Excel I/O support."
        ) from err

    write_excel_func(tables, path, na_rep)



from textwrap import dedent
from typing import Optional

import numpy as np
import pandas as pd
import pytest
from pint import DimensionalityError, UndefinedUnitError
from pytest import fixture, raises

import pdtable
from pdtable.units import PintUnitConverter
from ..demo.unit_converter import convert_this
from ..io.parsers.blocks import make_table
from ..proxy import UnitConversionNotDefinedError, MissingUnitConverterError


def test_demo_converter__works():
    # Converts single value
    assert convert_this(1, "m", "mm") == (1000, "mm")
    assert convert_this(0, "C", "K") == (273.15, "K")
    # Supports aliases
    assert convert_this(1000, "mm", "mètre") == (1, "m")
    # Returns NaN when given NaN
    new_val, _ = convert_this(np.nan, "m", "mm")
    assert np.isnan(new_val)
    # Converts array
    converted_vals, out_unit = convert_this(np.array([1, 42, np.nan]), "m", "mm")
    np.testing.assert_array_equal(converted_vals, np.array([1000, 42000, np.nan]))
    assert out_unit == "mm"
    # Converts to base unit by default
    assert convert_this(42_000, "mm") == (42, "m")
    # Fails when dimensionality error
    with raises(KeyError):
        convert_this(1, "m", "kg")
    # Fails when unknown units
    with raises(KeyError):
        convert_this(1, "m", "zonk")
    with raises(KeyError):
        convert_this(1, "gork", "m")
    with raises(KeyError):
        convert_this(1, "quxx")


def test_pint_converter__works():
    pint_converter = PintUnitConverter()
    # Converts single value
    assert pint_converter(1, "m", "mm") == (1000, "millimeter")
    assert pint_converter(0, "degC", "K") == (273.15, "kelvin")
    # Returns NaN when given NaN
    new_val, _ = pint_converter(np.nan, "m", "mm")
    assert np.isnan(new_val)
    # Converts array
    converted_vals, out_unit = pint_converter(np.array([1, 42]), "m", "mm")
    np.testing.assert_array_equal(converted_vals, np.array([1000, 42000]))
    assert out_unit == "millimeter"
    # Converts to base unit by default
    assert convert_this(42_000, "mm") == (42, "m")
    # Fails when dimensionality error
    with raises(DimensionalityError):
        # "C" means "Coulomb" in Pint's unit registry
        pint_converter(0, "C", "K")  # Can't convert Coulomb to kelvin
    # Fails when unknown units
    with raises(UndefinedUnitError):
        pint_converter(1, "m", "zonk")
    with raises(UndefinedUnitError):
        pint_converter(1, "gork", "m")
    with raises(UndefinedUnitError):
        pint_converter(1, "quxx")


class CustomUnitConverter(PintUnitConverter):
    def __init__(self):
        super().__init__()

    def __call__(self, value, from_unit, to_unit="__base_unit__"):
        # Let's say we think that "C" should mean "degrees Celsius" and not "Coulomb".
        custom_unit_symbols = {"C": "degC"}
        # Translate supplied unit symbols into pint unit symbols (if translation is defined)
        f = custom_unit_symbols.get(from_unit, from_unit)
        t = custom_unit_symbols.get(to_unit, to_unit)
        return super().__call__(value, f, t)


@fixture
def cuc():
    return CustomUnitConverter()


def test_custom_converter__works(cuc):
    # Pint units still work
    assert cuc(1, "m", "mm") == (1000, "millimeter")
    assert cuc(0, "degC", "K") == (273.15, "kelvin")
    # Units overridden in subclass work as intended
    assert cuc(0, "C", "K") == (273.15, "kelvin")


@fixture
def table_cells():
    return [
        [cell.strip() for cell in line.split(";")]
        for line in dedent(
            r"""
    **foo;
    all;
    diameter;mean_temp;depth;remark;measurement_date;
    mm;C;mm;text;datetime;
    42000;20;666;pretty cold;2020-10-09;
    1000;-;666;room temp;2020-10-09;
    """
        )
        .strip()
        .split("\n")
    ]


def test_convert_units__to_base_units(table_cells, cuc):
    t = make_table(table_cells).convert_units(to="base", converter=cuc)

    # Converted to base units
    np.testing.assert_array_equal(t["diameter"].values, np.array([42, 1]))
    assert t["diameter"].unit == "meter"
    np.testing.assert_array_equal(t["mean_temp"].values, np.array([293.15, np.nan]))
    assert t["mean_temp"].unit == "kelvin"
    np.testing.assert_array_equal(t["depth"].values, np.array([0.666, 0.666]))
    assert t["depth"].unit == "meter"
    # Columns with inconvertible units were skipped
    np.testing.assert_array_equal(t["remark"].values, np.array(["pretty cold", "room temp"]))
    assert t["remark"].unit == "text"
    assert all(x == pd.to_datetime("2020-10-09") for x in t["measurement_date"].values)
    assert t["measurement_date"].unit == "datetime"


def test_convert_units__list(table_cells, cuc):
    t = make_table(table_cells).convert_units(to=["m", "K", None, None, None], converter=cuc)

    # Conversion done on columns as requested
    np.testing.assert_array_equal(t["diameter"].values, np.array([42, 1]))
    assert t["diameter"].unit == "m"
    np.testing.assert_array_equal(t["mean_temp"].values, np.array([293.15, np.nan]))
    assert t["mean_temp"].unit == "K"

    # Column for which no conversion was requested stays unchanged
    np.testing.assert_array_equal(t["depth"].values, np.array([666, 666]))
    assert t["depth"].unit == "mm"


def test_convert_units__dict(table_cells, cuc):
    t = make_table(table_cells).convert_units(to={"diameter": "m", "mean_temp": "K"}, converter=cuc)

    # Conversion done on columns as requested
    np.testing.assert_array_equal(t["diameter"].values, np.array([42, 1]))
    assert t["diameter"].unit == "m"
    np.testing.assert_array_equal(t["mean_temp"].values, np.array([293.15, np.nan]))
    assert t["mean_temp"].unit == "K"

    # Column for which no conversion was requested stays unchanged
    np.testing.assert_array_equal(t["depth"].values, np.array([666, 666]))
    assert t["depth"].unit == "mm"


def test_convert_units__callable(table_cells, cuc):
    def to_units_fun(table_name: str) -> Optional[str]:
        return {"diameter": "m", "mean_temp": "K"}.get(table_name)

    t = make_table(table_cells).convert_units(to=to_units_fun, converter=cuc)

    # Conversion done on columns as requested
    np.testing.assert_array_equal(t["diameter"].values, np.array([42, 1]))
    assert t["diameter"].unit == "m"
    np.testing.assert_array_equal(t["mean_temp"].values, np.array([293.15, np.nan]))
    assert t["mean_temp"].unit == "K"

    # Column for which no conversion was requested stays unchanged
    np.testing.assert_array_equal(t["depth"].values, np.array([666, 666]))
    assert t["depth"].unit == "mm"


def test_convert_units__using_default_converter(table_cells, cuc):
    t_old_units = make_table(table_cells)

    assert pdtable.units.default_converter is None
    with raises(MissingUnitConverterError):
        # No default unit converter was set
        t_old_units.convert_units(to={"diameter": "m", "mean_temp": "K"})

    # Now set a default converter
    pdtable.units.default_converter = cuc
    # Do conversion; no explicitly specified unit converter; uses default
    t = t_old_units.convert_units(to={"diameter": "m", "mean_temp": "K"})

    # Conversion done on columns as requested
    np.testing.assert_array_equal(t["diameter"].values, np.array([42, 1]))
    assert t["diameter"].unit == "m"
    np.testing.assert_array_equal(t["mean_temp"].values, np.array([293.15, np.nan]))
    assert t["mean_temp"].unit == "K"

    # Column for which no conversion was requested stays unchanged
    np.testing.assert_array_equal(t["depth"].values, np.array([666, 666]))
    assert t["depth"].unit == "mm"


def test_convert_units__fails_on_inconvertible_unit(table_cells, cuc):
    t = make_table(table_cells)
    with pytest.raises(UnitConversionNotDefinedError):
        # Attempt to convert units of a datetime
        t.convert_units(to=[None, None, None, "m", None], converter=cuc)
    with pytest.raises(UnitConversionNotDefinedError):
        # Attempt to convert units of a text
        t.convert_units(to=[None, None, None, None, "m"], converter=cuc)



"""
    module containing auto-fixed input files from this directory
    need to be updated:
        1) when additional test-input are added
        2) when default ParseFixer are modified
        3) redo all.csv:
            rm -f all.csv
            cat *.csv > all.csv
    the file all.csv is not included here
"""
autoFixed = {
    "colname1.csv": """
        **farm_colname1;
        your_farm my_farm farms_galore
        species;-missing-;avg_weight
        text;-;kg
        chicken;2.0;3.0
        pig;4.0;39.0
        goat;4.0;-
        zybra;4.0;-
        cow;-;200.0
        goose;2.0;9.0
    """,
    "cols1.csv": """
        **farm_cols1;
        your_farm my_farm farms_galore
        species;num;flt;dt;flt_fixed_000;flt_fixed_001
        text;-;kg;datetime;kg;kg
        chicken;2.0;3.0;2020-07-01 00:00:00;3.21;1.0
        pig;4.0;39.0;2020-07-02 00:00:00;39.1;2.1
        goat;4.0;-;-;1.1;3.2
        zybra;4.0;-;-;2.1;4.3
        cow;-;200.0;-;200.2;5.4
        goose;2.0;9.0;-;9.1;6.5
        1234;-;-;-;7.11;7.6
    """,
    "cols2.csv": """
        **farm_cols2;
        your_farm my_farm farms_galore
        species;num;missing_fixed_000;dt;flt
        text;-;kg;datetime;kg
        chicken;2.0;3.0;2020-07-01 00:00:00;3.21
        pig;4.0;39.0;2020-07-02 00:00:00;39.1
        goat;4.0;-;-;1.1
        zybra;4.0;-;-;2.1
        cow;-;200.0;-;200.2
        goose;2.0;9.0;-;9.1
        1234;-;-;-;7.11
    """,
    "ex0.csv": """
        **farm_animals0;
        your_farm my_farm farms_galore
        species;n_legs;avg_weight
        text;-;kg
        chicken;2.0;3.0
        pig;4.0;30.0
        goat;4.0;-
        zybra;4.0;-
        cow;-;200.0
        goose;2.0;9.0
    """,
    "ex1.csv": """
        **farm_animals1;
        your_farm my_farm farms_galore
        species;n_legs;avg_weight
        text;-;kg
        chicken;2.0;3.0
        pig;4.0;39.0
        goat;4.0;-
        zybra;-;-
        cow;-;200.0
        goose;2.0;9.0
    """,
    "ex2.csv": """
        **farm_animals2;
        your_farm my_farm farms_galore
        species;n_legs;avg_weight
        text;-;kg
        chicken;2.0;-
        pig;4.0;39.0
        goat;4.0;-
        zybra;4.0;-
        cow;-;200.0
        goose;2.0;9.0
    """,
    "ex3.csv": """
        **farm_animals3;
        your_farm my_farm farms_galore
        species;n_legs;avg_weight
        text;-;kg
        chicken;2.0;-
        pig;4.0;39.0
        goat;4.0;-
        zybra;-;-
        cow;-;200.0
        goose;2.0;9.0
    """,
    "json1.csv": """
        **farm_json1;
        your_farm my_farm farms_galore
        species;dt;num;flt;dt2;encode
        text;datetime;-;kg;datetime;onoff
        "chicken";-;2.0;-;2020-07-01 00:00:00;1
        pig;2020-07-02 00:00:00;4.0;39.0;2020-07-02 00:00:00;0
        goat;-;-;-;-;0
        z'ybra;-;4.0;-;-;1
        'cow';-;-;200.0;-;1
        goose;-;2.0;9.0;-;0
    """,
    "row1.csv": """
        **farm_row1;
        your_farm my_farm farms_galore
        species;n_legs;avg_weight
        text;-;kg
        chicken;2.0;3.0
        pig;4.0;39.0
        goat;4.0;-
        zybra;-;-
        cow;-;200.0
        goose;2.0;9.0
    """,
    "row2.csv": """
        **farm_row2;
        your_farm my_farm farms_galore
        species;n_legs;avg_weight
        text;-;kg
        chicken;2.0;3.0
        pig;4.0;39.0
        goat;4.0;-
        zybra;4.0;3.0
        cow;-;200.0
        goose;2.0;9.0
    """,
    "types1.csv": """
        **farm_types1;
        your_farm my_farm farms_galore
        species;num;flt;log
        text;-;kg;onoff
        chicken;2.0;3.0;1
        pig;4.0;39.0;0
        goat;4.0;-;1
        zybra;4.0;-;0
        cow;-;200.0;1
        goose;2.0;9.0;0
    """,
    "types2.csv": """
        **farm_types2;
        your_farm my_farm farms_galore
        species;num;flt;dt
        text;-;kg;datetime
        chicken;2.0;3.0;2020-07-01 00:00:00
        pig;4.0;39.0;2020-07-02 00:00:00
        goat;4.0;-;-
        zybra;4.0;-;-
        cow;-;200.0;-
        goose;2.0;9.0;-
    """,
    "types3.csv": """
        **farm_types3;
        your_farm my_farm farms_galore
        species;dt;num;flt;log;dt2;flt2;log2
        text;datetime;-;kg;onoff;datetime;kg;onoff
        chicken;-;2.0;-;0;2020-07-01 00:00:00;-;0
        pig;2020-07-02 00:00:00;4.0;39.0;0;2020-07-02 00:00:00;39.0;0
        goat;-;-;-;1;-;-;1
        zybra;-;4.0;-;0;-;-;0
        cow;-;-;200.0;1;-;200.0;1
        goose;-;2.0;9.0;0;-;9.0;0
    """,
    "units1.csv": """
        **farm_units1;
        your_farm my_farm farms_galore
        species;num;flt
        text;-;kg
        chicken;2.0;3.0
        pig;4.0;39.0
        goat;4.0;-
        zybra;4.0;-
        cow;-;200.0
        goose;2.0;9.0
        1234;-;-
    """,
}



from typing import Union, Dict, List, Optional, Set, Callable, Sequence

import pandas as pd

from .frame import (
    TableDataFrame,
    get_table_info,
    is_table_dataframe,
    make_table_dataframe,
    set_units,
    add_column,
)
from .table_metadata import TableMetadata, ColumnMetadata, ComplementaryTableInfo
import pdtable  # for access to pdtable.units.default_converter

INCONVERTIBLE_UNIT_INDICATORS = ["text", "datetime", "onoff"]
UnitConverter = Callable[[float, str, str], float]
ColumnUnitDispatcher = Union[Sequence[str], Dict[str, str], Callable[[str], str]]


class UnitConversionNotDefinedError(ValueError):
    """Raised when a unit conversion is attempted on an inconvertible unit indicator"""

    pass


class Column:
    """
    Proxy for column in table

    TODO: Should we allow TableMetadata.unit to be None and then return "computed unit" as unit?
          Something like this would be needed if we are to change type of column
          via proxy interface. Alternative is to use "add_column()"
    """

    def __init__(self, df: TableDataFrame, name: str, table_info: ComplementaryTableInfo = None):
        self._name = name
        self._values = df[name]
        if not table_info:
            table_info = get_table_info(df)
        self._meta = table_info.columns[name]

    @property
    def name(self):
        return self._name

    @property
    def unit(self) -> str:
        """Column unit"""
        return self._meta.unit

    @unit.setter
    def unit(self, value: str):
        self._meta.unit = value

    @property
    def values(self):
        """Reference to the .array value of the underlying dataframe column

        Should not be edited directly.
        """
        return self._values.array

    @values.setter
    def values(self, values):
        self._values.update(pd.Series(values))

    def convert_units(self, to: Union[str, None], converter: UnitConverter):
        """Converts this column's units in place.

        Args:
            to:
                Can be any of:
                - '__base__': converts to the current unit's base unit
                - '__origin__': converts to the origin unit
                - other str: converts to this explicitly specified unit
                - None: no conversion
            converter:
                The converter.

        Returns:
            None

        """
        if to is None:
            # By convention, no unit conversion.
            return
        if to == self.unit:
            # It's already that unit. Nothing to do here.
            return
        if self.unit in INCONVERTIBLE_UNIT_INDICATORS:
            raise UnitConversionNotDefinedError(
                f"Unit conversion is not defined for unit '{self.unit}' of column '{self.name}'"
            )
        if to == "__origin__":
            # Convert to origin unit
            raise NotImplementedError  # TODO
        if to == "__base__":
            # Convert to base unit
            converted_values, to = converter(self.values, self.unit)
        else:
            # Convert to specified unit
            converted_values, _ = converter(self.values, self.unit, to)
        self.values = converted_values
        self.unit = to

    def to_numpy(self):
        """
        Value of column as numpy array. May require coercion and/or copying.
        """
        return self._values.to_numpy()

    def __repr__(self):
        return f"Column(name='{self.name}', unit='{self.unit}', values={self.values})"


class MissingUnitConverterError(ValueError):
    """Raised when a unit conversion cannot be done because no converter was made available."""

    pass


class Table:
    """
    A Table object is a facade for a backing TableDataFrame object.

    Can be created in two ways:
    1) From TableDataFrame object
       table = Table(tdf)
    2) From normal dataframe by including minimum metadata:
       table = Table(df, name='Foo')
       table = Table(df, ComplementaryTableInfo(name='Foo'))

    The .df property will return a dataframe subclass that retains all table information.
    To obtain a bare DataFrame object, use `pd.DataFrame(t.df)

    One possible performance issue is that since this facade does not have get messages
    from the backing object, we need to check the dataframe on each access.

    Note on performance: the Table class will check table_data against dataframe on each call.
    For situations where this is unacceptable for performance, use direct dataframe access methods.
    """

    def __init__(self, df: Union[None, TableDataFrame, pd.DataFrame] = None, **kwargs):
        if not (df is not None and is_table_dataframe(df)):
            # Creating a new table: initialize TableDataFrame
            df = make_table_dataframe(df if df is not None else pd.DataFrame(), **kwargs)
        elif kwargs:
            raise Exception(
                f"Got unexpected keyword arguments when creating Table object from "
                f"existing pandas table: {kwargs}"
            )
        self._df = df

    @property
    def df(self) -> TableDataFrame:
        """
        Return a pandas dataframe with all table information stored as
        metadata (a TableDataFrame object).

        This dataframe always exist and is the single source of truth for table data.
        The Table objects (this object) merely acts as a facade to allow simpler manipulation of
        associated metadata. It is consequently safe to simultaneously manipulate a Table object
        and the associated TableDataFrame object, as well as deleting the Table object.
        """
        return self._df

    @property
    def table_data(self) -> ComplementaryTableInfo:
        return get_table_info(self._df)

    @property
    def metadata(self) -> TableMetadata:
        return self.table_data.metadata

    @property
    def column_metadata(self) -> Dict[str, ColumnMetadata]:
        """
        Dictionary of column metadata objects

        It may be easier to use item access syntax (square brackets).
        Example:

        assert table.column_metadata['foo'].unit == table['foo'].unit
        """
        return self.table_data.columns

    # TODO: Rename columns -> column_names
    @property
    def column_names(self) -> List[str]:
        return self._df.columns.values.tolist()

    @property
    def column_proxies(self) -> List[Column]:
        df = self._df
        table_data = get_table_info(df)
        return [Column(df, name, table_info=table_data) for name in self.column_names]

    @property
    def units(self) -> List[str]:
        cols = self.column_metadata
        return [cols[name].unit for name in self.column_names]

    @property
    def name(self) -> str:
        return self.metadata.name

    @property
    def destinations(self) -> Set[str]:
        return self.metadata.destinations

    @units.setter
    def units(self, unit_values):
        set_units(self._df, unit_values)

    def get_row(self, index: int) -> List:
        # TODO call it 'row', and make it indexable, like iloc is?
        return self._df.iloc[index, :].values.tolist()

    def add_column(self, name: str, values, unit: Optional[str] = None, **kwargs):
        """
        Create or overwrite column.

        Example:
        table.add_column('some_column', 'm', range(4))

        An alternative is to use the item access syntax, e.g.:
        table['some_column'] = range(4)
        table['some_column'].unit = 'm'
        """
        add_column(self._df, name=name, values=values, unit=unit, **kwargs)

    def __iter__(self):
        table_data = self.table_data
        return (Column(self._df, name, table_info=table_data) for name in table_data.columns.keys())

    def __getitem__(self, name: str):
        """Get column proxy for existing column"""
        return Column(self._df, name)

    def __setitem__(self, name: str, values):
        """"Create or update column value

        If column is created, unit is derived from dtype.
        value: any value accepted for dataframe column creation.
        """
        add_column(self._df, name, values)

    def as_dataframe_with_annotated_column_names(self) -> pd.DataFrame:
        """
        Returns a dataframe with units included in column names
        """
        df = self._df.copy()
        cm = self.column_metadata
        df.columns = [f"{c} [{cm[c].unit}]" for c in df.columns]
        return df

    def __repr__(self):
        m = self.metadata
        # TODO __repr__ shouldn't display the dataframe's index. Could also display units on their own line.  # noqa
        return (
            f"**{m.name}\n{', '.join(s for s in m.destinations)}"
            f"\n{self.as_dataframe_with_annotated_column_names()}"
        )

    def __str__(self):
        return repr(self)

    def __metadata_comp_key(self):
        """Metadata comparison key, for use in __eq__"""
        return self.name, self.metadata.destinations, self.column_names, self.units

    def equals(self, other):
        """Checks whether the other Table has the same header and data as this one.

        Checks whether the other Table has the same name, destinations, column names, column units,
        and data values as this one. Origin is ignored (tables can be equal regardless of where they
        came from).

        Number equality does not take data type into account, in keeping with the StarTable
        convention that a number is just a number. For example, float 10.0 equals int 10.

        This is implemented as a method rather than as the __eq__() magic method because a later
        implementation of __hash__() attempting to live by the rule
        "a == b implies hash(a) == hash(b)" would have to do a deep dive in the table data
        to force equal hashes on equal numbers, even when they naturally have different hashes due
        to having different data types.
        """
        if isinstance(other, self.__class__):
            self_key = self.__metadata_comp_key()
            other_key = other.__metadata_comp_key()
            return self_key == other_key and _df_elements_all_equal_or_same(self._df, other._df)
            # Had to implement this custom equality checker for DataFrames because,
            # as of pandas 1.1.0, stupid pandas.DataFrame.equals return False when elements have
            # different dtypes e.g. 10 and 10.0 are considered 'not equal'. In StarTable, a number
            # is just a number, and no such distinction should be made between data types.
        return False

    def convert_units(self, to: ColumnUnitDispatcher, converter: UnitConverter = None) -> "Table":
        """Applies unit conversion to columns, modifying table in-place

        Returns a new table with converted units. The desired new unit are specified by
        column.

        How to do a conversion from unit X to unit Y is determined by the supplied unit
        converter.

        The converter is also responsible for deciding what unit is considered the
        "base unit" of unit X.  Depending on your application and favourite unit system, the base
        unit of 'mm' could be 'm', 'foot', 'furlong', or some other unit of dimension length.
        
        Args:
            to:
                Specifies to what units to convert which columns. Can be:
                - 'base': Converts all columns to their respective base units. Columns with
                  inconvertible unit indicators are skipped.
                - 'origin': (not yet implemented!) Converts all columns to their respective origin
                  units. Columns with inconvertible unit indicators are skipped.
                - A dictionary of {column_name: target_unit}. Superfluous column names are ignored.
                - A callable with one argument: column name. Must return the target unit, or None
                  if no unit conversion is to be done.
                - A list specifying the target unit of each column by position. (A None element
                  implies no conversion for that column.) This is discouraged in production, as
                  there is no check that the units are applied to the right columns by column name;
                  but it is a useful shorthand during experimentation.

                Individual columns' target units can be specified as any of:
                - '__base__': the column's current unit's base unit
                - '__origin__': the column's origin unit
                - other str: explicitly specified unit
                - None: keep the column's current unit; do no conversion.

            converter:
                A callable that converts values from one unit to another. The converter must
                fulfill the following specification:
                - Takes three positional arguments:
                    0. value to be converted
                    1. from_unit
                    2. Optional: to_unit. If the caller does not specify to_unit, the converter
                       should assume that the target unit is from_unit's base unit. If this
                       assumption is not implemented in the converter, then calls to
                       convert_units(to='base') will fail.
                - Accepts units of type returned by the ColumnUnitDispatcher.
                - Returns a tuple of two elements:
                    0.  The value with unit conversion applied. Should be NaN if input value is NaN.
                    1.  The converter's string representation of the target unit. This may differ
                        from to_unit. For example, a pint-based converter called with to_unit="mm"
                        would return "millimeter".

        Returns:
            A new Table with converted units.

        """
        default_converter = pdtable.units.default_converter
        if converter is None:
            if default_converter is None:
                raise MissingUnitConverterError(
                    "No converter or default converter was specified.",
                    converter,
                    default_converter,
                )
            converter = default_converter

        new_table = Table(self.df.copy())

        if to == "origin":
            for col in new_table.column_proxies:
                if col.unit in INCONVERTIBLE_UNIT_INDICATORS:
                    # Skip this column
                    continue
                col.convert_units("__origin__", converter)

        elif to == "base":
            # Convert all columns to their respective base units
            for col in new_table.column_proxies:
                if col.unit in INCONVERTIBLE_UNIT_INDICATORS:
                    # Skip this column
                    continue
                col.convert_units("__base__", converter)

        elif isinstance(to, Sequence):
            if len(to) != len(self.column_proxies):
                raise ValueError(
                    "Unequal number of columns and of 'to' units", len(self.column_proxies), len(to)
                )
            for col, to_unit in zip(new_table.column_proxies, to):
                if to_unit is not None:
                    col.convert_units(to_unit, converter)

        elif isinstance(to, Dict):
            for column in new_table.column_proxies:
                to_unit = to.get(column.name)
                if to_unit is not None:
                    column.convert_units(to[column.name], converter)

        elif isinstance(to, Callable):
            for column in new_table.column_proxies:
                to_unit = to(column.name)
                if to_unit is not None:
                    column.convert_units(to_unit, converter)

        else:
            raise TypeError("Column unit dispatcher of unexpected type.", type(to), to)

        return new_table


def _equal_or_same(a, b):
    """Returns True if both values are equal or 'the same thing' (e.g. NaN's).

    Note that np.nan != float('nan') != pd.NA etc., so we collapse all of these
    missing value things using pd.isna()
    """
    return a == b or a is b or (pd.isna(a) and pd.isna(b))


def _df_elements(df):
    """Yields all the values in the data frame serially."""
    return (x for row in df.itertuples() for x in row)


def _df_elements_all_equal_or_same(df1, df2):
    """Returns True if all corresponding elements are equal or 'the same' in both data frames."""
    try:
        return all(_equal_or_same(x1, x2) for x1, x2 in zip(_df_elements(df1), _df_elements(df2)))
    except Exception:
        # If the comparison can't be made, then clearly they aren't the same
        return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""A setuptools based setup module.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

from setuptools import setup, find_packages
from os import path
import re

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()


def find_version(*file_paths):
    """
    Reads version from a file. Version must be specified explicitly in the file as:
    __version__ = "<the version number string>"
    """
    version_file = open(path.join(*file_paths), "r").read()
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


setup(
    name="pdtable",
    version=find_version("pdtable", "__init__.py"),
    description="Reads and writes data stored in StarTable format; and stores table data in"
    "memory as a Pandas data frame for easy manipulation.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/startable/pdtable/",
    author="Jean-François Corbett",
    author_email="jeaco@orsted.dk",
    license="BSD-3-Clause",
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3",
    ],
    # What does your project relate to?
    keywords="startable data-structure file-format table dataframe json",
    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=["contrib", "doc", "docs", "test", "tests"]),
    python_requires=">=3.7",
    install_requires=["numpy", "pandas"],
    tests_require=["pytest", "openpyxl", "pint"],
    # Same for developer dependencies
    # extras_require={
    #     'dev': ['check-manifest'],
    #     'test': ['coverage'],
    # },
)



# flake8: noqa

__version__ = "0.0.2"

CSV_SEP = ";"  # User can overwrite this default

from .proxy import Table
from .frame import TableDataFrame
from .table_metadata import TableMetadata, TableOrigin
from .auxiliary import Directive, MetadataBlock
from .store import TableBundle, BlockType, BlockIterator
from .io import ParseFixer
from .io import read_csv, write_csv
from .io import read_excel, write_excel
from .units.pint import PintUnitConverter



from pathlib import Path
from pytest import raises

from ... import Table
from ...io import read_excel
from ...store import BlockType


def test_read_excel():

    # Prepare the expected tables.
    # Note: deliberately not testing datetime columns due to upstream bug in openpyxl:
    # timestamps are off by one microsecond.
    # https://foss.heptapod.net/openpyxl/openpyxl/-/issues/1493

    t0 = Table(name="places_to_go")
    t0["place"] = ["home", "work", "beach", "wonderland"]
    t0.add_column("distance", list(range(3)) + [float("nan")], "km")
    t0.add_column("is_hot", [True, False, True, False], "onoff")

    t1 = Table(name="spelling_numbers")
    t1.add_column("number", [1, 6, 42], "-")
    t1.add_column("spelling", ["one", "six", "forty-two"], "text")

    expected_tables = [t0, t1]

    # Read tables from file
    blocks = read_excel(Path(__file__).parent / "input" / "foo.xlsx")
    tables_read = [block for (block_type, block) in blocks if block_type == BlockType.TABLE]
    assert len(expected_tables) == len(tables_read)

    # Assert read tables are equal to the expected ones
    for te, tr in zip(expected_tables, tables_read):
        assert te.equals(tr)

    # test_read_excel__from_stream
    with open(Path(__file__).parent / "input" / "foo.xlsx", "rb") as fh:
        blocks = read_excel(fh)
        tables_read_stream = [
            block for (block_type, block) in blocks if block_type == BlockType.TABLE
        ]
        assert len(expected_tables) == len(tables_read_stream)

    # raises exception if not binary stream
    with raises(Exception):
        with open(Path(__file__).parent / "input" / "foo.xlsx", "r") as fh:  # text stream!
            blocks = read_excel(fh)
            tables_read_stream = [
                block for (block_type, block) in blocks if block_type == BlockType.TABLE
            ]


def test_read_excel__applies_filter():

    # Make a filter
    def is_table_about_spelling(block_type: BlockType, block_name: str) -> bool:
        return block_type == BlockType.TABLE and "spelling" in block_name

    # Read blocks from file
    blocks = list(
        read_excel(Path(__file__).parent / "input" / "foo.xlsx", filter=is_table_about_spelling)
    )

    # Assert only that one table block was parsed
    assert len(blocks) == 1
    table: Table = blocks[0][1]
    assert table.name == "spelling_numbers"



import datetime as dt

import numpy as np
import pandas as pd
import pytest
from numpy.testing import assert_array_equal
from pytest import raises

from pdtable.io.parsers.columns import (
    normalize_if_str,
    is_missing_data_marker,
    _parse_onoff_column,
    _parse_float_column,
    _parse_datetime_column,
    _parse_text_column,
    parse_column,
)


def test_normalize_if_str():
    assert normalize_if_str(" NoRmALiZe me\t") == "normalize me"
    assert normalize_if_str("already normalized") == "already normalized"
    assert normalize_if_str(42) == 42
    assert normalize_if_str(None) is None


@pytest.mark.parametrize(
    "x,out",
    [
        ("-", True),
        ("NaN", True),
        ("nan", True),
        ("NAN", True),
        ("nAn", True),
        ("Non!", False),
        (None, False),
    ],
)
def test_is_missing_data_marker(x, out):
    assert is_missing_data_marker(x) == out


def test__parse_text_column():
    strings = ["foo", "bart", ""]
    max_str_length = max(len(s) for s in strings)
    col = _parse_text_column(strings)
    assert_array_equal(col, np.array(["foo", "bart", ""]))
    assert col.dtype == np.dtype(f"<U{max_str_length}")  # that's how numpy does dtype for strings


def test__parse_onoff_column():
    col = _parse_onoff_column([0, 1, False, True, "0", "1", " 0\t", "1 \n "])
    assert_array_equal(col, np.array([False, True, False, True, False, True, False, True,]))
    assert col.dtype == bool


def test__parse_onoff_column__panics_on_illegal_value():
    illegal_values = ["-", 2, -1]
    for x in illegal_values:
        with raises(ValueError):
            _parse_onoff_column([x])


def test__parse_float_column():
    col = _parse_float_column([-1, 42, "-1", "42", "-", "NaN", "nan"])
    assert_array_equal(col, np.array([-1, 42, -1, 42, np.nan, np.nan, np.nan]))
    assert col.dtype == float


def test__parse_float_column__panics_on_illegal_value():
    illegal_values = ["foo", "", None]
    for x in illegal_values:
        with raises(ValueError):
            _parse_float_column([x])


def test__parse_datetime_column():
    col = _parse_datetime_column(["2020-08-11", dt.datetime(2020, 8, 11, 11, 40)])
    assert_array_equal(
        col, np.array([pd.to_datetime("2020-08-11"), pd.to_datetime("2020-08-11 11:40")])
    )

    col = _parse_datetime_column(["-", "nan"])
    assert all(v is pd.NaT for v in col)


@pytest.mark.parametrize(
    "unit_indicator,values,expected",
    [
        ("text", ["yes", "", "no"], np.array(["yes", "", "no"])),
        (
            "onoff",
            [0, 1, False, True, "0", "1", " 0\t", "1 \n "],
            np.array([False, True, False, True, False, True, False, True,]),
        ),
        (
            "float",
            [-1, 42, "-1", "42", "-", "NaN", "nan"],
            np.array([-1, 42, -1, 42, np.nan, np.nan, np.nan]),
        ),
        (
            "datetime",
            ["2020-08-11", dt.datetime(2020, 8, 11, 11, 40)],
            np.array([pd.to_datetime("2020-08-11"), pd.to_datetime("2020-08-11 11:40")]),
        ),
    ],
)
def test__parse_column(unit_indicator, values, expected):
    assert_array_equal(parse_column(unit_indicator, values), expected)



"""A courtesy pint-based unit converter."""

from typing import Union, Tuple


class PintUnitConverter:
    """A thin callable-class wrapper around Pint.

    Subclass to customize. Or use as-is if Pint's unit registry does it for you.

    This class should be instantiated only once and used as a singleton, to avoid creating multiple,
    unrelated instances of the pint unit registry.
    """

    def __init__(self):

        try:
            import pint
        except ImportError as err:
            raise ImportError(
                "Unable to import 'pint'. "
                "Please install 'pint' to use this pint-based unit converter."
            ) from err
        # Initialize unit registry once, keep it around for multiple calls
        self.ureg = pint.UnitRegistry()

    def __call__(
        self,
        value: float,
        from_unit: Union[str],
        to_unit: Union[str] = "__base_unit__",
    ) -> Tuple[float, str]:
        """Converts value from one unit to another unit.

        Both units must be defined in Pint's unit registry.

        If to_unit is not specified, it defaults to from_unit's base unit.
        For example, converting 2 'cm' will return 0.02 'meter'.

        Args:
            value: Magnitude of quantity to convert
            from_unit: Unit of quantity to convert.
            to_unit: Unit to which to convert.  If None (default), converts to base unit.

        Returns:
            The converted quantity's magnitude and unit. Note that if to_unit is specified,
            then the returned unit is pint's string representation of to_unit.
            If, however, to_unit is not specified, then the returned unit is pint's string
            representation of from_unit's base unit.

        """
        if str(to_unit) == str(from_unit):
            # Null conversion
            return value, str(from_unit)
        elif to_unit == "__base_unit__":
            # Convert to base unit
            converted_quantity = self.ureg.Quantity(value, from_unit).to_base_units()
        else:
            # Convert to specified unit
            converted_quantity = self.ureg.Quantity(value, from_unit).to(to_unit)

        return converted_quantity.magnitude, str(converted_quantity.units)
